"""Implements functionality unique to the M81 Source Modules."""

import math
from datetime import datetime
from warnings import warn

from lakeshore.xip_instrument import RegisterBase
from lakeshore.ssm_base_module import SSMSystemModuleQuestionableRegister, BaseModule
from lakeshore.requires_firmware_version import requires_firmware_version


class SSMSystemSourceModuleOperationRegister(RegisterBase):
    """Class object representing the operation status register of a source module"""

    bit_names = [
        "v_limit",
        "i_limit",
        "sweeping"
    ]

    def __init__(
            self,
            v_limit,
            i_limit,
            sweeping):
        self.v_limit = v_limit
        self.i_limit = i_limit
        self.sweeping = sweeping


# pylint: disable=R0904
class SourceModule(BaseModule):
    """Class for interaction with a specific source channel of the M81 instrument"""

    def get_multiple(self, *data_sources):
        r"""This function is deprecated. Use fetch_multiple() instead."""

        return self.fetch_multiple(*data_sources)

    def fetch_multiple(self, *data_sources):
        r"""Gets a list of the latest values corresponding to the input data sources for this module, and returns them
        as soon as possible.

        Args:
            data_sources (SSMSystemDataSourceMnemonic or str): Variable length list of data sources.

        Returns:
            Tuple of values corresponding to the given data sources for this module
        """

        elements = [(data_source, self.module_number) for data_source in data_sources]
        return self.device.fetch_multiple(*elements)

    def get_name(self):
        """Returns the user-settable name of the module"""

        return self.device.query(f'SOURce{self.module_number}:NAME?').strip('\"')

    def set_name(self, new_name):
        """Set the name of the module"""

        self.device.command(f'SOURce{self.module_number}:NAME "{new_name}"')

    def get_notes(self):
        """Returns the user-settable notes of the module"""

        return self.device.query(f'SOURce{self.module_number}:NOTes?').strip('\"')

    def set_notes(self, new_note):
        """Set the notes of the module"""

        self.device.command(f'SOURce{self.module_number}:NOTes "{new_note}"')

    def get_model(self):
        """Returns the model of the module (i.e. BCS-10)"""

        return self.device.query(f'SOURce{self.module_number}:MODel?').strip('\"')

    def get_serial(self):
        """Returns the serial number of the module (i.e. LSA1234)"""

        return self.device.query(f'SOURce{self.module_number}:SERial?').strip('\"')

    def get_hw_version(self):
        """Returns the hardware version of the module"""

        return int(self.device.query(f'SOURce{self.module_number}:HWVersion?'))

    def get_self_cal_status(self):
        """Returns the status of the last self calibration of the module"""

        return self.device.query(f'SOURce{self.module_number}:SCALibration:STATus?')

    def run_self_cal(self):
        """Run a self calibration for the module"""

        self.device.command(f'SOURce{self.module_number}:SCALibration:RUN')

    def reset_self_cal(self):
        """Restore factory self calibration for the module"""

        self.device.command(f'SOURce{self.module_number}:SCALibration:RESet')

    def get_enable_state(self):
        """Returns the output state of the module"""

        return bool(int(self.device.query(f'SOURce{self.module_number}:STATe?')))

    def set_enable_state(self, state):
        """Sets the enable state of the module

            Args:
                state (bool):
                    The new output state
        """

        self.device.command(f'SOURce{self.module_number}:STATe {str(int(state))}')

    def enable(self):
        """Sets the enable state of the module to True"""

        self.set_enable_state(True)

    def disable(self):
        """Sets the enable state of the module to False"""

        self.set_enable_state(False)

    def get_excitation_mode(self):
        """Returns the excitation mode of the module. 'CURRENT' or 'VOLTAGE'."""

        return self.device.ExcitationType(self.device.query(f'SOURce{self.module_number}:FUNCtion:MODE?'))

    def set_excitation_mode(self, excitation_mode):
        """Sets the excitation mode of the module

            Args:
                excitation_mode (SSMSystem.ExcitationType):
                    The new excitation mode ('CURRENT' or 'VOLTAGE')
        """

        if isinstance(excitation_mode, self.device.ExcitationType):
            mode = excitation_mode.name
        else:
            mode = excitation_mode

        self.device.command(f'SOURce{self.module_number}:FUNCtion:MODE {mode}')

    def go_to_current_mode(self):
        """Sets the excitation mode of the module to 'CURRENT'"""

        self.set_excitation_mode(self.device.ExcitationType.CURRENT)

    def go_to_voltage_mode(self):
        """Sets the excitation mode of the module to 'VOLTAGE'"""

        self.set_excitation_mode(self.device.ExcitationType.VOLTAGE)

    def get_shape(self):
        """Returns the signal shape of the module. 'DC' or 'SINUSOID'."""

        return self.device.query(f'SOURce{self.module_number}:FUNCtion:SHAPe?')

    def set_shape(self, shape):
        """Sets the signal shape of the module

            Args:
                shape (str):
                    The new signal shape ('DC', 'SINUSOID', 'TRIANGLE', 'SQUARE')
        """

        self.device.command(f'SOURce{self.module_number}:FUNCtion:SHAPe {shape}')

    def get_frequency(self):
        """Returns the excitation frequency of the module"""

        return float(self.device.query(f'SOURce{self.module_number}:FREQuency?'))

    def set_frequency(self, frequency):
        """Sets the excitation frequency of the module

            Args:
                frequency (float):
                    The new excitation frequency
        """
        self.device.command(f'SOURce{self.module_number}:FREQuency {str(frequency)}')

    def get_sync_state(self):
        """Returns whether the source channel synchronization feature is engaged

            If true, this channel will ignore its own frequency, and instead track the frequency of the synchronization source.
            If false, this channel will generate its own frequency.
        """

        return bool(int(self.device.query(f'SOURce{self.module_number}:SYNChronize:STATe?')))

    def get_sync_source(self):
        """Returns the channel used for frequency synchronization"""

        return self.device.query(f'SOURce{self.module_number}:SYNChronize:SOURce?')

    def get_sync_phase_shift(self):
        """Returns the phase shift applied between the synchronization source and this channel"""

        return float(self.device.query(f'SOURce{self.module_number}:SYNChronize:PHASe?'))

    def configure_sync(self, source, phase_shift, enable_sync=True):
        """Configure the source channel synchronization feature

            Args:
                source (str):
                    The channel used for synchronization ('S1', 'S2', 'S3', or 'RIN').
                    This channel will follow the frequency set for the specified channel.

                phase_shift (float):
                    The phase shift applied between the synchronization source and this channel in degrees.

                enable_sync (bool):
                    If true, this channel will ignore its own frequency, and instead track the frequency of the synchronization source.
                    If false, this channel will generate its own frequency.
        """
        self.device.command(f'SOURce{self.module_number}:SYNChronize:SOURce {source}')
        self.device.command(f'SOURce{self.module_number}:SYNChronize:PHASe {str(phase_shift)}')
        self.device.command(f'SOURce{self.module_number}:SYNChronize:STATe {str(int(enable_sync))}')

    def get_duty(self):
        """Returns the duty cycle of the module"""

        return float(self.device.query(f'SOURce{self.module_number}:DCYCle?'))

    def set_duty(self, duty):
        """Sets the duty cycle of the module

            Args:
                duty (float):
                    The new duty cycle
        """

        self.device.command(f'SOURce{self.module_number}:DCYCle {str(duty)}')

    def get_coupling(self):
        """Returns the coupling type of the module. 'AC' or 'DC'."""

        return self.device.query(f'SOURce{self.module_number}:COUPling?')

    def set_coupling(self, coupling):
        """Sets the coupling of the module

            Args:
                coupling (str):
                    The new coupling type ('AC', or 'DC')
        """
        self.device.command(f'SOURce{self.module_number}:COUPling {coupling}')

    def use_ac_coupling(self):
        """Sets the coupling type of the module to 'AC'"""

        self.set_coupling('AC')

    def use_dc_coupling(self):
        """Sets the coupling type of the module to 'DC'"""

        self.set_coupling('DC')

    def get_guard_state(self):
        """Returns the guard state of the module"""

        return bool(int(self.device.query(f'SOURce{self.module_number}:GUARd?')))

    def set_guard_state(self, guard_state):
        """Sets the guard state of the module

            Args:
                guard_state (bool):
                    The new guard state (True to enable guards, False to disable guards)
        """

        self.device.command(f'SOURce{self.module_number}:GUARd {str(int(guard_state))}')

    def enable_guards(self):
        """Sets the guard state of the module to True"""

        self.set_guard_state(True)

    def disable_guards(self):
        """Sets the guard state of the module to False"""

        self.set_guard_state(False)

    def get_cmr_source(self):
        """Returns the Common Mode Reduction (CMR) source. 'INTernal', or 'EXTernal'."""

        return self.device.query(f'SOURce{self.module_number}:CMR:SOURce?')

    def set_cmr_source(self, cmr_source):
        """Sets the Common Mode Reduction (CMR) source.

            Args:
                cmr_source (str):
                    The new CMR source ('INTernal', or 'EXTernal')
        """

        self.device.command(f'SOURce{self.module_number}:CMR:SOURce {cmr_source}')

    def get_cmr_state(self):
        """Returns the Common Mode Reduction (CMR) state of the module"""

        return bool(int(self.device.query(f'SOURce{self.module_number}:CMR:STATe?')))

    def set_cmr_state(self, cmr_state):
        """Sets the Common Mode Reduction (CMR) state of the module

            Args:
                cmr_state (bool):
                    The new CMR state (True to enable CMR, False to disable CMR)
        """

        self.device.command(f'SOURce{self.module_number}:CMR:STATe {str(int(cmr_state))}')

    def enable_cmr(self):
        """Sets the CMR state of the module to True"""

        self.set_cmr_state(True)

    def disable_cmr(self):
        """Sets the CMR state of the module to False"""

        self.set_cmr_state(False)

    def configure_cmr(self, cmr_source, cmr_state=True):
        """Configure Common Mode Reduction (CMR)

            Args:
                cmr_source (str):
                    The new CMR source ('INTernal', or 'EXTernal')

                cmr_state (bool):
                    The new CMR state (True to enable CMR, False to disable CMR)
        """

        self.set_cmr_source(cmr_source)
        self.set_cmr_state(cmr_state)

    def get_current_range(self):
        """Returns the present current range of the module in Amps"""

        return float(self.device.query(f'SOURce{self.module_number}:CURRent:RANGe?'))

    def get_i_range(self):
        """
        Returns the present current range of the module in Amps
        
        .. deprecated:: 1.5.4
           Use get_current_range instead.
        """

        warn("The get_i_range method is deprecated, use get_current_range instead", DeprecationWarning)
        return self.get_current_range()

    def get_current_ac_range(self):
        """Returns the present AC current range of the module in Amps"""

        return float(self.device.query(f'SOURce{self.module_number}:CURRent:RANGe:AC?'))

    def get_i_ac_range(self):
        """
        Returns the present AC current range of the module in Amps
        
        .. deprecated:: 1.5.4
           Use get_current_ac_range instead.
        """

        warn("The get_i_ac_range method is deprecated, use get_current_ac_range instead", DeprecationWarning)
        return self.get_current_ac_range()

    def get_current_dc_range(self):
        """Returns the present DC current range of the module in Amps"""

        return float(self.device.query(f'SOURce{self.module_number}:CURRent:RANGe:DC?'))

    def get_i_dc_range(self):
        """
        Returns the present DC current range of the module in Amps
        
        .. deprecated:: 1.5.4
           Use get_current_dc_range instead.
        """

        warn("The get_i_dc_range method is deprecated, use get_current_dc_range instead", DeprecationWarning)
        return self.get_current_dc_range()

    def get_current_autorange_status(self):
        """Returns whether automatic selection of the current range is enabled for this module"""

        return bool(int(self.device.query(f'SOURce{self.module_number}:CURRent:RANGe:AUTO?')))

    def get_i_autorange_status(self):
        """
        Returns whether automatic selection of the current range is enabled for this module
        
        .. deprecated:: 1.5.4
           Use get_current_autorange_status instead.
        """

        warn("The get_i_autorange_status method is deprecated, use get_current_autorange_status instead", DeprecationWarning)
        return self.get_current_autorange_status()

    def configure_current_range(self, autorange, max_level=None, max_ac_level=None, max_dc_level=None):
        """Sets up current ranging for this module

            Args:
                autorange (bool):
                    True to enable automatic range selection. False for manual ranging.

                max_level (float):
                    The largest current that needs to be sourced.

                max_ac_level (float):
                    The largest AC current that needs to be sourced. Separate AC and DC ranges are only available on some modules.

                max_dc_level (float):
                    The largest DC current that needs to be sourced. Separate AC and DC ranges are only available on some modules.
        """

        if autorange:
            if max_level is not None or max_ac_level is not None or max_dc_level is not None:
                raise ValueError('If autorange is selected, a manual range cannot be specified.')

            self.device.command(f'SOURce{self.module_number}:CURRent:RANGe:AUTO 1')
        else:
            if max_level is not None:
                if max_ac_level is not None or max_dc_level is not None:
                    raise ValueError('Either a single range, or separate AC and DC ranges can be supplied, not both.')

                self.device.command(f'SOURce{self.module_number}:CURRent:RANGe {str(max_level)}')
            else:
                if max_ac_level is not None:
                    self.device.command(f'SOURce{self.module_number}:CURRent:RANGe:AC {str(max_ac_level)}')
                if max_dc_level is not None:
                    self.device.command(f'SOURce{self.module_number}:CURRent:RANGe:DC {str(max_dc_level)}')

    def configure_i_range(self, autorange, max_level=None, max_ac_level=None, max_dc_level=None):
        """
        Sets up current ranging for this module

            Args:
                autorange (bool):
                    True to enable automatic range selection. False for manual ranging.

                max_level (float):
                    The largest current that needs to be sourced.

                max_ac_level (float):
                    The largest AC current that needs to be sourced. Separate AC and DC ranges are only available on some modules.

                max_dc_level (float):
                    The largest DC current that needs to be sourced. Separate AC and DC ranges are only available on some modules.
        
        .. deprecated:: 1.5.4
           Use configure_current_range instead.
        """

        warn("The configure_i_range method is deprecated, use configure_current_range instead", DeprecationWarning)
        self.configure_current_range(autorange, max_level, max_ac_level, max_dc_level)

    def get_current_amplitude(self):
        """Returns the current amplitude for the module in Amps"""

        return float(self.device.query(f'SOURce{self.module_number}:CURRent:LEVel:AMPLitude?'))

    def get_i_amplitude(self):
        """
        Returns the current amplitude for the module in Amps
        
        .. deprecated:: 1.5.4
           Use get_current_amplitude instead.
        """

        warn("The get_i_amplitude method is deprecated, use get_current_amplitude instead", DeprecationWarning)
        return self.get_current_amplitude()

    def set_current_amplitude(self, amplitude):
        """Sets the current amplitude for the module

            Args:
                amplitude (float):
                    The new current amplitude in Amps
        """
        self.device.command(f'SOURce{self.module_number}:CURRent:LEVel:AMPLitude {str(amplitude)}')

    def set_i_amplitude(self, amplitude):
        """
        Sets the current amplitude for the module

            Args:
                amplitude (float):
                    The new current amplitude in Amps
        
        .. deprecated:: 1.5.4
           Use set_current_amplitude instead.
        """

        warn("The set_i_amplitude method is deprecated, use set_current_amplitude instead", DeprecationWarning)
        self.set_current_amplitude(amplitude)

    def get_current_offset(self):
        """Returns the current offset for the module in Amps"""

        return float(self.device.query(f'SOURce{self.module_number}:CURRent:LEVel:OFFSet?'))

    def get_i_offset(self):
        """
        Returns the current offset for the module in Amps
        
        .. deprecated:: 1.5.4
           Use get_current_offset instead.
        """

        warn("The get_i_offset method is deprecated, use get_current_offset instead", DeprecationWarning)
        return self.get_current_offset()

    def set_current_offset(self, offset):
        """Sets the current offset for the module

            Args:
                offset (float):
                    The new current offset in Amps
        """

        self.device.command(f'SOURce{self.module_number}:CURRent:LEVel:OFFSet {str(offset)}')

    def set_i_offset(self, offset):
        """
        Sets the current offset for the module

            Args:
                offset (float):
                    The new current offset in Amps
        
        .. deprecated:: 1.5.4
           Use set_current_offset instead.
        """

        warn("The set_i_offset method is deprecated, use set_current_offset instead", DeprecationWarning)
        self.set_current_offset(offset)

    def apply_dc_current(self, level, output_enable=True):
        """Apply DC current

            Args:
                level (float):
                    DC current level in Amps

                output_enable (bool):
                    Turns the module output on if true; off if false
        """

        if not output_enable:
            self.disable()

        self.set_excitation_mode('CURRent')
        self.set_shape('DC')
        self.set_current_amplitude(level)

        if output_enable:
            self.enable()

    def apply_ac_current(self, frequency, amplitude, offset=0.0, output_enable=True):
        """Apply AC current

            Args:
                frequency (float):
                    Excitation frequency in Hz

                amplitude (float):
                    Current amplitude in Amps

                offset (float):
                    Current offset in Amps

                output_enable (bool):
                    Turns the module output on if true; off if false
        """

        if not output_enable:
            self.disable()

        self.set_excitation_mode('CURRent')
        self.set_frequency(frequency)
        self.set_shape('SINusoid')
        self.set_current_amplitude(amplitude)
        self.set_current_offset(offset)

        if output_enable:
            self.enable()

    def get_current_limit(self):
        """Returns the current limit enforced by the module in Amps"""

        return float(self.device.query(f'SOURce{self.module_number}:CURRent:PROTection?'))

    def get_i_limit(self):
        """
        Returns the current limit enforced by the module in Amps
        
        .. deprecated:: 1.5.4
           Use get_current_limit instead.
        """

        warn("The get_i_limit method is deprecated, use get_current_limit instead", DeprecationWarning)
        return self.get_current_limit()

    def set_current_limit(self, current_limit):
        """Sets the current limit enforced by the module

            Args:
                current_limit (float):
                    The new limit to apply in Amps
        """
        self.device.command(f'SOURce{self.module_number}:CURRent:PROTection {str(current_limit)}')

    def set_i_limit(self, i_limit):
        """
        Sets the current limit enforced by the module

            Args:
                i_limit (float):
                    The new limit to apply in Amps
        
        .. deprecated:: 1.5.4
           Use set_current_limit instead.
        """

        warn("The set_i_limit method is deprecated, use set_current_limit instead", DeprecationWarning)
        self.set_current_limit(i_limit)

    def get_current_limit_status(self):
        """Returns whether the current limit circuitry is presently engaged and limiting the current sourced by the module"""

        return bool(int(self.device.query(f'SOURce{self.module_number}:CURRent:PROTection:TRIPped?')))

    def get_i_limit_status(self):
        """
        Returns whether the current limit circuitry is presently engaged and limiting the current sourced by the module
        
        .. deprecated:: 1.5.4
           Use get_current_limit_status instead.
        """

        warn("The get_i_limit_status method is deprecated, use get_current_limit_status instead", DeprecationWarning)
        return self.get_current_limit_status()

    def get_voltage_range(self):
        """Returns the present voltage range of the module in Volts"""

        return float(self.device.query(f'SOURce{self.module_number}:VOLTage:RANGe?'))

    def get_voltage_ac_range(self):
        """Returns the present AC voltage range of the module in Volts"""

        return float(self.device.query(f'SOURce{self.module_number}:VOLTage:RANGe:AC?'))

    def get_voltage_dc_range(self):
        """Returns the present DC voltage range of the module in Volts"""

        return float(self.device.query(f'SOURce{self.module_number}:VOLTage:RANGe:DC?'))

    def get_voltage_autorange_status(self):
        """Returns whether automatic selection of the voltage range is enabled for this module"""

        return bool(int(self.device.query(f'SOURce{self.module_number}:VOLTage:RANGe:AUTO?')))

    def configure_voltage_range(self, autorange, max_level=None, max_ac_level=None, max_dc_level=None):
        """Sets up voltage ranging for this module

            Args:
                autorange (bool):
                    True to enable automatic range selection. False for manual ranging.

                max_level (float):
                    The largest voltage that needs to be sourced.

                max_ac_level (float):
                    The largest AC voltage that needs to be sourced. Separate AC and DC ranges are only available on some modules.

                max_dc_level (float):
                    The largest DC voltage that needs to be sourced. Separate AC and DC ranges are only available on some modules.
        """

        if autorange:
            if max_level is not None or max_ac_level is not None or max_dc_level is not None:
                raise ValueError('If autorange is selected, a manual range cannot be specified.')

            self.device.command(f'SOURce{self.module_number}:VOLTage:RANGe:AUTO 1')
        else:
            if max_level is not None:
                if max_ac_level is not None or max_dc_level is not None:
                    raise ValueError('Either a single range, or separate AC and DC ranges can be supplied, not both.')

                self.device.command(f'SOURce{self.module_number}:VOLTage:RANGe {str(max_level)}')
            else:
                if max_ac_level is not None:
                    self.device.command(f'SOURce{self.module_number}:VOLTage:RANGe:AC {str(max_ac_level)}')
                if max_dc_level is not None:
                    self.device.command(f'SOURce{self.module_number}:VOLTage:RANGe:DC {str(max_dc_level)}')

    def get_voltage_amplitude(self):
        """Returns the voltage amplitude for the module in Volts"""

        return float(self.device.query(f'SOURce{self.module_number}:VOLTage:LEVel:AMPLitude?'))

    def set_voltage_amplitude(self, amplitude):
        """Sets the voltage amplitude for the module

            Args:
                amplitude (float):
                    The new voltage amplitude in Volts
        """

        self.device.command(f'SOURce{self.module_number}:VOLTage:LEVel:AMPLitude {str(amplitude)}')

    def get_voltage_offset(self):
        """Returns the voltage offset for the module in Volts"""

        return float(self.device.query(f'SOURce{self.module_number}:VOLTage:LEVel:OFFSet?'))

    def set_voltage_offset(self, offset):
        """Sets the voltage offset for the module

            Args:
                offset (float):
                    The new voltage offset in Volts
        """

        self.device.command(f'SOURce{self.module_number}:VOLTage:LEVel:OFFSet {str(offset)}')

    def apply_dc_voltage(self, level, output_enable=True):
        """Apply DC voltage

            Args:
                level (float):
                    DC voltage level in Volts

                output_enable (bool):
                    Turns the module output on if true; off if false
        """

        if not output_enable:
            self.disable()

        self.set_excitation_mode('VOLTage')
        self.set_shape('DC')
        self.set_voltage_amplitude(level)

        if output_enable:
            self.enable()

    def apply_ac_voltage(self, frequency, amplitude, offset=0.0, output_enable=True):
        """Apply AC voltage

            Args:
                frequency (float):
                    Excitation frequency in Hz

                amplitude (float):
                    Voltage amplitude in Volts

                offset (float):
                    Voltage offset in Volts

                output_enable (bool):
                    Turns the module output on if true; off if false
        """

        if not output_enable:
            self.disable()

        self.set_excitation_mode(self.device.ExcitationType.VOLTAGE)
        self.set_frequency(frequency)
        self.set_shape('SINusoid')
        self.set_voltage_amplitude(amplitude)
        self.set_voltage_offset(offset)

        if output_enable:
            self.enable()

    def get_voltage_limit(self):
        """Returns the voltage limit enforced by the module in Volts"""

        return float(self.device.query(f'SOURce{self.module_number}:VOLTage:PROTection?'))

    def set_voltage_limit(self, voltage_limit):
        """Sets the voltage limit enforced by the module

            Args:
                voltage_limit (float):
                    The new limit to apply in Volts
        """

        self.device.command(f'SOURce{self.module_number}:VOLTage:PROTection {str(voltage_limit)}')

    def get_voltage_limit_status(self):
        """Returns whether the voltage limit circuitry is presently engaged and limiting the voltage at the output of the module"""

        return bool(int(self.device.query(f'SOURce{self.module_number}:VOLTage:PROTection:TRIPped?')))

    def get_present_questionable_status(self):
        """Returns the names of the questionable status register bits and their values"""

        response = self.device.query(f'STATus:QUEStionable:SOURce{self.module_number}:CONDition?', check_errors=False)
        status_register = SSMSystemModuleQuestionableRegister.from_integer(response)

        return status_register

    def get_questionable_events(self):
        """Returns the names of questionable event status register bits that are currently high.
        The event register is latching and values are reset when queried."""

        response = self.device.query(f'STATus:QUEStionable:SOURce{self.module_number}:EVENt?', check_errors=False)
        status_register = SSMSystemModuleQuestionableRegister.from_integer(response)

        return status_register

    def get_questionable_event_enable_mask(self):
        """Returns the names of the questionable event enable register bits and their values.
        These values determine which questionable bits propagate to the questionable event register."""

        response = self.device.query(f'STATus:QUEStionable:SOURce{self.module_number}:ENABle?', check_errors=False)
        status_register = SSMSystemModuleQuestionableRegister.from_integer(response)

        return status_register

    def set_questionable_event_enable_mask(self, register_mask):
        """Configures the values of the questionable event enable register bits.
        These values determine which questionable bits propagate to the questionable event register.

            Args:
                register_mask ([Instrument]QuestionableRegister):
                    An instrument specific QuestionableRegister class object with all bits configured true or false.
        """

        integer_representation = register_mask.to_integer()
        self.device.command(f'STATus:QUEStionable:SOURce{self.module_number}:ENABle {integer_representation}', check_errors=False)

    def get_present_operation_status(self):
        """Returns the names of the operation status register bits and their values"""

        response = self.device.query(f'STATus:OPERation:SOURce{self.module_number}:CONDition?', check_errors=False)
        status_register = SSMSystemSourceModuleOperationRegister.from_integer(response)

        return status_register

    def get_operation_events(self):
        """Returns the names of operation event status register bits that are currently high.
        The event register is latching and values are reset when queried."""

        response = self.device.query(f'STATus:OPERation:SOURce{self.module_number}:EVENt?', check_errors=False)
        status_register = SSMSystemSourceModuleOperationRegister.from_integer(response)

        return status_register

    def get_operation_event_enable_mask(self):
        """Returns the names of the operation event enable register bits and their values.
        These values determine which operation bits propagate to the operation event register."""

        response = self.device.query(f'STATus:OPERation:SOURce{self.module_number}:ENABle?', check_errors=False)
        status_register = SSMSystemSourceModuleOperationRegister.from_integer(response)

        return status_register

    def set_operation_event_enable_mask(self, register_mask):
        """Configures the values of the operation event enable register bits.
        These values determine which operation bits propagate to the operation event register.

            Args:
                register_mask ([Instrument]OperationRegister):
                    An instrument specific OperationRegister class object with all bits configured true or false.
        """

        integer_representation = register_mask.to_integer()
        self.device.command(f'STATus:OPERation:SOURce{self.module_number}:ENABle {integer_representation}', check_errors=False)

    def get_identify_state(self):
        """Returns the identification state for the given pod."""
        response = bool(int(self.device.query(f'SOURce{self.module_number}:IDENtify?', check_errors=False)))
        return response

    def set_identify_state(self, state):
        """Configures the identification state for the given pod.

            Args:
                state (bool):
                    The desired state for the LED, 1 for identify, 0 for normal state
        """
        self.device.command(f'SOURce{self.module_number}:IDENtify {int(state)}', check_errors=False)

    def get_dark_mode_state(self):
        """Returns the dark mode state for the given pod"""
        response = self.device.query(f'SOURce{self.module_number}:DMODe?', cherk_errors=False)
        return response

    def set_dark_mode_state(self, state):
        """Configures the dark mode state for the given pod.

            Args:
                state (bool):
                    The desired operation for the LED, 1 for normal mode, 0 for dark mode
        """
        self.device.command(f'SOURce{self.module_number}:DMODe {state}', check_errors=False)

    def get_voltage_output_limit_high(self):
        """Returns the present voltage high output limit."""
        response = float(self.device.query(f'SOURce{self.module_number}:VOLTage:LIMit:HIGH?', cherk_errors=False))
        return response

    def set_voltage_output_limit_high(self, limit):
        """Configures the high voltage output limit.
        The voltage output limits are software defined limits preventing the user from entering an output which could
        potentially damage the module's load.
        When the shape is not DC, the limit is applied to the sum of the offset and amplitude.
        The high voltage output limit is bounded between -10 V and 10 V, and must be greater than the low voltage output limit.

            Args:
                limit (float):
                    The desired high output limit.
        """
        self.device.command(f'SOURce{self.module_number}:VOLTage:LIMit:HIGH {limit}', check_errors=False)

    def get_voltage_output_limit_low(self):
        """Returns the present voltage low output limit."""
        response = float(self.device.query(f'SOURce{self.module_number}:VOLTage:LIMit:LOW?', cherk_errors=False))
        return response

    def set_voltage_output_limit_low(self, limit):
        """Configures the low voltage output limit.
        The voltage output limits are software defined limits preventing the user from entering an output which could
        potentially damage the module's load.
        When the shape is not DC, the limit is applied to the sum of the offset and amplitude.
        The low voltage output limit is bounded between -10 V and 10 V, and must be less than the high voltage output limit.

            Args:
                limit (float):
                    The desired low voltage output limit.
        """
        self.device.command(f'SOURce{self.module_number}:VOLTage:LIMit:LOW {limit}', check_errors=False)

    def get_current_output_limit_high(self):
        """Returns the present current high output limit."""
        response = float(self.device.query(f'SOURce{self.module_number}:CURRent:LIMit:HIGH?', cherk_errors=False))
        return response

    def set_current_output_limit_high(self, limit):
        """Configures the high current output limit.
        The current output limits are software defined limits preventing the user from entering an output which could
        potentially damage the module's load.
        When the shape is not DC, the limit is applied to the sum of the offset and amplitude.
        The high current output limit is bounded between -10 V and 10 V, and must be greater than the low current output limit.

            Args:
                limit (float):
                    The desired high output limit.
        """
        self.device.command(f'SOURce{self.module_number}:CURRent:LIMit:HIGH {str(limit)}', check_errors=False)

    def get_current_output_limit_low(self):
        """Returns the present current low output limit."""
        response = float(self.device.query(f'SOURce{self.module_number}:CURRent:LIMit:LOW?', cherk_errors=False))
        return response

    def set_disable_on_compliance(self, disable_on_compliance):
        """Configures the module for disable on compliance.
        When disable on compliance is turned on, the module will disable output when in compliance.
        Otherwise, the module will continue to output, even when in compliance.

            Args:
                disable_on_compliance (bool):
                    1 for the module to disable when in compliance; 0 for the module to remain enabled, even in compliance
        """
        self.device.command(f'SOURce{self.module_number}:DOCompliance {int(disable_on_compliance)}', check_errors=False)

    def get_disable_on_compliance(self):
        """Returns the present state of disable on compliance."""
        response = bool(self.device.query(f'SOURce{self.module_number}:DOCompliance?', check_errors=False))
        return response

    def set_current_output_limit_low(self, limit):
        """Configures the low current output limit.
        The current output limits are software defined limits preventing the user from entering an output which could
        potentially damage the module's load.
        When the shape is not DC, the limit is applied to the sum of the offset and amplitude.
        The low current output limit is bounded between -10 V and 10 V, and must be less than the high current output limit.

            Args:
                limit (float):
                    The desired low current output limit.
        """
        self.device.command(f'SOURce{self.module_number}:CURRent:LIMit:LOW {str(limit)}', check_errors=False)

    def reset_settings(self):
        """Resets the settings for the specified module to their power on defaults.
        """

        self.device.command(f'SOURce{self.module_number}:PRESet')

    def unload(self):
        """Unloads the specified module.
        """

        self.device.command(f'SOURce{self.module_number}:UNLoad')

    def get_load_state(self):
        """Returns the loaded state for the specified module.
        """

        response = bool(int(self.device.query(f'SOURce{self.module_number}:LOAD?')))
        return response

    def get_self_cal_datetime(self):
        """Returns the self calibration date and time for the specified module.
        """

        response = self.device.query(f'SOURce{self.module_number}:SCALibration:DATE?').split(',')
        return datetime(int(response[0]), int(response[1]), int(response[2]), int(response[3]), int(response[4]), int(response[5]))

    def get_self_cal_temperature(self):
        """Returns the self calibration temperature for the specified module.
        """

        response = float(self.device.query(f'SOURce{self.module_number}:SCALibration:TEMP?'))
        return response

    @requires_firmware_version('1.7.0')
    def get_source_sweep_step_size(self, sweep_type):
        """
        Returns the step size of the source sweep for the specified module.
        Step size is a calculated parameter derived from the relevant sweep type's span and points.

        Args:
            sweep_type (SourceSweepType):
                The type of sweep for which to return the step size.
        """

        response = float(self.device.query(f'SOURce{self.module_number}:{sweep_type}:STEP?'))
        return response

    @requires_firmware_version('1.7.0')
    def get_source_sweep_time(self):
        """
        Returns the overall runtime of the source sweep for the specified module in seconds.
        Sweep time is a calculated parameter derived from the dwell time and number of points.
        """

        response = float(self.device.query(f'SOURce{self.module_number}:SWEep:TIME?'))
        return response

    @requires_firmware_version('1.7.0')
    def abort_source_sweep(self):
        """Aborts all in progress source sweep for the specified module."""

        self.device.command(f'SOURce{self.module_number}:SWEep:ABORt')

    @requires_firmware_version('1.7.0')
    def get_source_sweep_state(self):
        """Returns the state of the source sweep on the specified module."""

        response = bool(int(self.device.query(f'SOURce{self.module_number}:SWEep:STATus?')))
        return response

    @requires_firmware_version('1.7.0')
    def set_sweep_configuration(self, sweep_settings):
        """
        Configures a source sweep for the specified module.

        Args:
            sweep_settings (SourceSweepSettings):
                The configuration for a specific sweep on the specified module.
        """

        self.device.command(f'SOURce{self.module_number}:SWEep:DWELl {sweep_settings.dwell}')
        self.device.command(f'SOURce{self.module_number}:SWEep:POINts {sweep_settings.points}')
        self.device.command(f'SOURce{self.module_number}:SWEep:SPACing {sweep_settings.spacing}')
        self.device.command(f'SOURce{self.module_number}:SWEep:DIRection {sweep_settings.direction}')
        self.device.command(f'SOURce{self.module_number}:SWEep:DIRection:RTRip {int(sweep_settings.round_trip)}')
        self.device.command(f'SOURce{self.module_number}:{sweep_settings.sweep_type}:MODE SWEep')
        self.device.command(f'SOURce{self.module_number}:{sweep_settings.sweep_type}:STARt {sweep_settings.start}')
        self.device.command(f'SOURce{self.module_number}:{sweep_settings.sweep_type}:STOP {sweep_settings.stop}')

    @requires_firmware_version('1.7.0')
    def get_sweep_configuration(self, sweep_type):
        """
        Returns a SourceSweepSettings of the present sweep configuration for the specified module.

        Args:
            sweep_type (SourceSweepType):
                The sweep type for which to return the sweep settings.
        """

        return self.device.SourceSweepSettings(
            sweep_type,
            float(self.device.query(f'SOURce{self.module_number}:{sweep_type}:STARt?')),
            float(self.device.query(f'SOURce{self.module_number}:{sweep_type}:STOP?')),
            int(self.device.query(f'SOURce{self.module_number}:SWEep:POINts?')),
            float(self.device.query(f'SOURce{self.module_number}:SWEep:DWELl?')),
            self.device.query(f'SOURce{self.module_number}:SWEep:DIRection?'),
            bool(int(self.device.query(f'SOURce{self.module_number}:SWEep:DIRection:RTRip?'))),
            self.device.query(f'SOURce{self.module_number}:SWEep:SPACing?'))

    @requires_firmware_version('1.7.0')
    def disable_all_sweeping(self):
        """Disables all source signals that support sweeping on the specified module."""

        for sweep_type in self.device.SourceSweepType:
            self.device.command(f'SOURce{self.module_number}:{sweep_type}:MODE FIXED')

    @requires_firmware_version('1.7.0')
    def disable_sweeping(self, sweep_type):
        """
        Disables the sweeping of the specified sweep type on the specified module.

        Args:
            sweep_type (SourceSweepType):
                The type of sweep to disable.
        """

        self.device.command(f'SOURce{self.module_number}:{sweep_type}:MODE FIXED')

    def set_voltage_ramp_configuration(self, stop_amplitude, start_amplitude=None, slew_rate=1.0):
        """
        Sets up a parameter sweep that ramps the voltage output to the desired amplitude using the smallest
        possible step size.

        Args:
            stop_amplitude (float):
                The voltage amplitude of the output when the ramp completes.

            start_amplitude (float):
                The voltage amplitude of the output when the ramp starts. Default is the present amplitude setting.

            slew_rate (float):
                The rate in volts per second to ramp the output. Default is 1 volt per second.
        """
        # Use the present voltage amplitude as the starting point if no start is specified
        if start_amplitude is None:
            start_amplitude = self.get_voltage_amplitude()

        ramp_total_time = abs(stop_amplitude - start_amplitude) / abs(slew_rate)

        # Determine the shortest dwell time that can be used without exceeding the maximum number of points
        dwell_time = math.ceil(ramp_total_time / (self.device.min_sweep_dwell * self.device.max_sweep_points)) * self.device.min_sweep_dwell
        num_points = round(ramp_total_time / dwell_time)

        sweep_config = self.device.SourceSweepSettings(sweep_type=self.device.SourceSweepType.VOLTAGE_AMPLITUDE,
                                                       start=start_amplitude,
                                                       stop=stop_amplitude,
                                                       points=num_points,
                                                       dwell=dwell_time,
                                                       direction=self.device.SourceSweepSettings.Direction.UP,
                                                       round_trip=False)
        self.set_sweep_configuration(sweep_config)

    def set_current_ramp_configuration(self, stop_amplitude, start_amplitude=None, slew_rate=0.001):
        """
        Sets up a parameter sweep that ramps the current output to the desired amplitude using the smallest
        possible step size.

        Args:
            stop_amplitude (float):
                The current amplitude of the output when the ramp completes.

            start_amplitude (float):
                The current amplitude of the output when the ramp starts. Default is the present amplitude setting.

            slew_rate (float):
                The rate in amps per second to ramp the output. Default is 1 mA per second.
        """
        # Use the present voltage amplitude as the starting point if no start is specified
        if start_amplitude is None:
            start_amplitude = self.get_current_amplitude()

        ramp_total_time = abs(stop_amplitude - start_amplitude) / abs(slew_rate)

        # Determine the shortest dwell time that can be used without exceeding the maximum number of points
        dwell_time = math.ceil(ramp_total_time / (self.device.min_sweep_dwell * self.device.max_sweep_points)) * self.device.min_sweep_dwell
        num_points = round(ramp_total_time / dwell_time)

        sweep_config = self.device.SourceSweepSettings(sweep_type=self.device.SourceSweepType.CURRENT_AMPLITUDE,
                                                       start=start_amplitude,
                                                       stop=stop_amplitude,
                                                       points=num_points,
                                                       dwell=dwell_time,
                                                       direction=self.device.SourceSweepSettings.Direction.UP,
                                                       round_trip=False)
        self.set_sweep_configuration(sweep_config)



"""Implements functionality unique to settings profiles"""

import json


class SettingsProfiles:
    """Class for interaction with settings profiles"""

    def __init__(self, device):
        self.device = device

    def get_summary(self, name):
        """Returns a list containing a profile's description and module models.

        Args:
            name (str): Name of the profile to query.
        """

        response = self.device.query(f'PROFile:SUMMary? "{name}"')

        return [element.replace('"', '').strip() for element in response.split(',')]

    def create(self, name, description=''):
        """Create a new profile using the present instrument configuration.

        Args:
            name (str): Unique name to give the profile.
            description (str): Optional description of the profile.
        """

        self.device.command(f'PROFile:CREAte "{name}", "{description}"')

    def get_list(self):
        """Returns a list of the saved profile names."""

        response = self.device.query('PROFile:LIST?')
        if not response:
            return []

        return [profile.strip('"') for profile in response.split(',')]

    def get_description(self, name):
        """Returns a profile's description.

        Args:
            name (str): Name of the profile to get the description for.
        """

        return self.device.query(f'PROFile:DESCription? "{name}"').strip('"')

    def set_description(self, name, description):
        """Sets a profile's description. Any existing description will be overwritten.

        Args:
            name (str): Name of the profile to get the description for.
            description (str): The new description of the profile.
        """

        self.device.command(f'PROFile:DESCription "{name}","{description}"')

    def get_json(self, name):
        """Returns a JSON object of a given profile.

        Args:
            name (str): Name of the profile.
        """

        response = self.device.query(f'PROFile:JSON? "{name}"')
        json_string = response.strip('"').replace('""', '"')
        return json.loads(json_string)

    def rename(self, name, new_name):
        """Rename a profile. New name must be unique.

        Args:
            name (str): The name of the profile to rename.
            new_name (str): The new name of the profile.
        """

        self.device.command(f'PROFile:REName "{name}","{new_name}"')

    def update(self, name):
        """Update a profile with the present instrument configuration.

        Args:
            name (str): The name of the profile to update.
        """

        self.device.command(f'PROFile:UPDate "{name}"')

    def get_valid_for_restore(self, name):
        """Returns if a profile is valid to restore.

        Args:
            name (str): The name of the profile to validate.
        """

        response = self.device.query(f'PROFile:RESTore:VALid? "{name}"')
        return bool(response)

    def restore(self, name):
        """Restore a profile.

        Args:
            name (str): The name of the profile to restore.
        """

        self.device.command(f'PROFile:RESTore "{name}"')

    def delete(self, name):
        """Delete a profile

        Args:
            name (str): The name of the profile to delete.
        """

        self.device.command(f'PROFile:DELete "{name}"')

    def delete_all(self):
        """Delete all profiles."""

        self.device.command('PROFile:DELete:ALL')



"""This module implements a parent class that contains basic functionality for communicating with Lake Shore instruments."""
import logging
import select
import socket
from threading import Lock
from time import sleep

import serial
from serial.tools.list_ports import comports


class InstrumentException(Exception):
    """Names a new type of exception specific to general instrument connectivity."""


class RegisterBase:
    """Base class of the status register classes"""

    bit_names = []

    def __str__(self):
        return str(vars(self))

    def to_integer(self):
        """Translates the register object to an integer representation value"""

        integer_representation = 0

        # Add up the boolean values of a list of named instrument states
        # while being careful to account for unnamed entries in the register bit names list
        for count, bit_name in enumerate(self.bit_names):

            if bit_name:
                integer_representation += int(getattr(self, bit_name)) << count

        return integer_representation

    @classmethod
    def from_integer(cls, integer_representation):
        """Creates the register object from an integer representation value"""

        # Create a dictionary to temporarily store the bit states
        bit_states = {}

        # Assign the boolean value of each bit in the integer to the corresponding status register bit name
        for count, bit_name in enumerate(cls.bit_names):
            if bit_name:
                mask = 0b1 << count
                bit_states[bit_name] = bool(int(integer_representation) & mask)

        return cls(**bit_states)


class GenericInstrument:
    """Parent class that implements functionality to connect to generic instruments"""

    vid_pid = []
    logger = logging.getLogger(__name__)

    def __init__(self, serial_number, com_port, baud_rate, data_bits, stop_bits, parity, flow_control,
                 handshaking, timeout, ip_address, tcp_port, connection=None):
        # Initialize values common to all instruments
        self.device_serial = None
        self.device_tcp = None
        self.dut_lock = Lock()
        self.serial_number = None
        self.option_card_serial = None

        # Raise an error if serial and TCP parameters are passed. Otherwise connect to the instrument using one of them.
        if ip_address is not None:
            if com_port is not None:
                raise ValueError("Two different connection methods provided.")

            self.connect_tcp(ip_address, tcp_port, timeout)
        else:
            if connection is None:
                self.connect_usb(serial_number, com_port, baud_rate, data_bits, stop_bits, parity,
                                 timeout, handshaking, flow_control)
            else:
                self.device_serial = connection

        # Query the instrument identification information and store it in variables
        try:
            idn_response = self._get_identity()
            self.firmware_version = idn_response[3]
            serial_string = idn_response[2].split("/")
            self.serial_number = serial_string[0]
            if len(serial_string) == 2:
                self.option_card_serial = serial_string[1]
            self.model_number = idn_response[1]
        except InstrumentException:
            print('Instrument found but unable to communicate. Please check interface settings on the instrument.')
            raise

        # Check to make sure the serial number matches what was provided if connecting over TCP
        if ip_address is not None and serial_number is not None and serial_number != self.serial_number:
            raise InstrumentException("Instrument found but the serial number does not match. " +
                                      "serial number provided is " + serial_number +
                                      ", serial number found is " + self.serial_number)

    def __del__(self):
        if self.device_serial is not None:
            self.device_serial.close()
        if self.device_tcp is not None:
            self.device_tcp.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.__del__()

    def command(self, command_string):
        """Send a command to the instrument

            Args:
                command_string (str):
                    A serial command
        """

        # Query the instrument over serial. If serial is not configured, use TCP.
        with self.dut_lock:
            # Send command to the instrument over serial. If serial is not configured, send it over TCP.
            if self.device_serial is not None:
                self._usb_command(command_string)
            elif self.device_tcp is not None:
                self._tcp_command(command_string)
            else:
                raise InstrumentException("No connections configured")

            self.logger.info('Sent command to %s: %s', self.serial_number, command_string)

    def query(self, query_string):
        """Send a query to the instrument and return the response

            Args:
                query_string (str):
                    A serial query ending in a question mark

            Returns:
               The instrument query response as a string.

        """

        # Query the instrument over serial. If serial is not configured, use TCP.
        with self.dut_lock:
            if self.device_serial is not None:
                response = self._usb_query(query_string)
            elif self.device_tcp is not None:
                response = self._tcp_query(query_string)
            else:
                raise InstrumentException("No connections configured")

            self.logger.info('Sent query to %s: %s', self.serial_number, query_string)
            self.logger.info('Received response from %s: %s', self.serial_number, response)

        return response

    def connect_tcp(self, ip_address, tcp_port, timeout):
        """Establishes a TCP connection with the instrument on the specified IP address"""

        self.device_tcp = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.device_tcp.settimeout(timeout)
        self.device_tcp.connect((ip_address, tcp_port))

        # Send the instrument a line break, wait 100ms, and clear the input buffer so that
        # any leftover communications from a prior session don't gum up the works.
        self.device_tcp.send(b'\n')
        sleep(0.1)
        while True:
            read_objects, _, _ = select.select([self.device_tcp], [], [], 0.0)
            if not read_objects:
                break
            for read_object in read_objects:
                read_object.recv(1)

    def disconnect_tcp(self):
        """Disconnect the TCP connection"""

        self.device_tcp.close()
        self.device_tcp = None

    def connect_usb(self, serial_number=None, com_port=None, baud_rate=None, data_bits=None,
                    stop_bits=None, parity=None, timeout=None, handshaking=None, flow_control=None):
        """Establish a serial USB connection"""

        # Scan the ports for devices matching the VID and PID combos of the instrument
        for port in comports():
            if (port.vid, port.pid) in self.vid_pid:
                # If the com port argument is passed, check for a match
                if port.device == com_port or com_port is None:
                    # If the serial number argument is passed, check for a match
                    if port.serial_number == serial_number or serial_number is None:
                        # Establish a connection with device using the instrument's serial communications parameters
                        self.device_serial = serial.Serial(port.device,
                                                           baudrate=baud_rate,
                                                           bytesize=data_bits,
                                                           stopbits=stop_bits,
                                                           xonxoff=handshaking,
                                                           timeout=timeout,
                                                           parity=parity,
                                                           rtscts=flow_control)

                        # Send the instrument a line break, wait 100ms, and clear the input buffer so that
                        # any leftover communications from a prior session don't gum up the works
                        self.device_serial.write(b'\n')
                        sleep(0.1)
                        self.device_serial.reset_input_buffer()

                        break
        else:
            if com_port is None and serial_number is None:
                raise InstrumentException("No serial connections found")

            raise InstrumentException(
                "No serial connections found with a matching COM port and/or matching serial number")

    def disconnect_usb(self):
        """Disconnect the USB connection"""

        self.device_serial.close()
        self.device_serial = None

    def _tcp_command(self, command):
        """Send a command over the TCP connection"""

        self.device_tcp.send(command.encode('utf-8') + b'\n')

    def _tcp_query(self, query):
        """Query over the TCP connection"""

        self._tcp_command(query)

        total_response = ""

        # Continuously receive data from the buffer until a line break
        while True:

            # Receive the data and raise an error on timeout
            try:
                response = self.device_tcp.recv(4096).decode('utf-8')
            except socket.timeout as ex:
                raise InstrumentException("Connection timed out") from ex

            # Add received information to the response
            total_response += response

            # Return the response once it ends with a line break
            if total_response.endswith("\r\n"):
                return total_response.rstrip()

    def _usb_command(self, command):
        """Send a command over the serial USB connection"""

        self.device_serial.write(command.encode('ascii') + b'\n')

    def _usb_query(self, query):
        """Query over the serial USB connection"""

        self._usb_command(query)
        response = self.device_serial.read_until(b'\r\n').decode('ascii')

        # If nothing is returned, raise a timeout error.
        if not response:
            raise InstrumentException("Communication timed out")

        return response.rstrip()

    def _get_identity(self):
        return self.query('*IDN?').split(',')



# -*- coding: utf-8 -*-
"""Implements functionality unique to the Lake Shore Model 240 channel modules"""
from enum import IntEnum
import serial

from .generic_instrument import GenericInstrument


class Model240Units(IntEnum):
    """Enumerations that specify temperature units"""
    KELVIN = 1
    CELSIUS = 2
    SENSOR = 3
    FAHRENHEIT = 4


class Model240CurveFormat(IntEnum):
    """Enumerations that specify temperature sensor curve format units"""
    VOLTS_PER_KELVIN = 2
    OHMS_PER_KELVIN = 3
    LOG_OHMS_PER_KELVIN = 4


class Model240Coefficients(IntEnum):
    """Enumerations that specify a positive or negative coefficient"""
    NEGATIVE = 1
    POSITIVE = 2


class Model240SensorTypes(IntEnum):
    """Enumerations specify types of temperature sensors"""
    DIODE = 1
    PLATINUM_RTD = 2
    NTC_RTD = 3


class Model240BrightnessLevel(IntEnum):
    """Enumerations for the screen brightness levels"""
    OFF = 0
    LOW = 25
    MED_LOW = 50
    MED_HIGH = 75
    HIGH = 100


class Model240TemperatureCoefficient(IntEnum):
    """Enumerations specify positive/negative temperature sensor curve coefficients"""
    NEGATIVE = 1
    POSITIVE = 2


class Model240InputRange(IntEnum):
    """Enumerations to specify the input range when autorange is off"""
    RANGE_DIODE = 0
    RANGE_PTRTD_1_KIL_OHMS = 0
    RANGE_NTCRTD_10_OHMS = 0
    RANGE_NTCRTD_30_OHMS = 1
    RANGE_NTCRTD_100_OHMS = 2
    RANGE_NTCRTD_1_KIL_OHMS = 4
    RANGE_NTCRTD_3_KIL_OHMS = 5
    RANGE_NTCRTD_10_KIL_OHMS = 6
    RANGE_NTCRTD_30_KIL_OHMS = 7
    RANGE_NTCRTD_100_KIL_OHMS = 8


class Model240CurveHeader:
    """A class that configures the user curve header and corresponding parameters"""

    def __init__(self, curve_name, serial_number, curve_data_format, temperature_limit, coefficient):
        """Constructor for CurveHeader class

            Args:
                curve_name (str):
                    * Specifies curve name (limit of 15 characters)

                serial_number (str):
                    * Specifies curve serial number (limit of 10 characters)

                curve_data_format (Model240CurveFormat):
                    * Specifies the curve data format

                temperature_limit (float):
                    * Specifies the curve temperature limit in Kelvin

                coefficient (Model240TemperatureCoefficient):
                    * Specifies the curve temperature coefficient

        """

        self.curve_name = curve_name
        self.serial_number = serial_number
        self.curve_data_format = curve_data_format
        self.temperature_limit = temperature_limit
        self.coefficient = coefficient


class Model240InputParameter:
    """Class used to retrieve and set an input channel's parameters and initial settings."""

    def __init__(self, sensor, auto_range_enable, current_reversal_enable, units, input_enable, input_range=None):
        """The constructor for InputParameter class.

            Args:

                sensor (Model240SensorTypes):
                    * Specifies the type of sensor configured at the input
                    * Member of the Model240SensorTypes IntEnum class

                auto_range_enable (bool):
                    Specifies if autoranging is enabled

                current_reversal_enable (bool):
                    * Specifies channel current reversal
                    * Current reversal is used to remove thermal EMF errors on resistive sensors.
                    * Always False if channel is a diode
                    * False = OFF, True = ON

                units (Model240Units):
                    * Member of the Model240Units IntEnum class
                    * Specifies the preferred units parameter.

                input_enable (bool):
                    Specifies whether the channel is disabled or enabled

                input_range (Model240InputRange):
                    Specifies channel range when autorange is off

        """

        self.sensor_type = sensor
        self.temperature_unit = units
        self.auto_range_enable = auto_range_enable
        self.current_reversal_enable = current_reversal_enable
        self.input_enable = input_enable
        self.input_range = input_range


class Model240ProfiSlot:
    """Class used to configure and retrieve data for given PROFIBUS slot."""

    def __init__(self, channel, temp_unit):
        """The constructor for Model240ProfiSlot class.

            Args:
                channel (int):
                    Specifies which slot to configure (1-8)

                temp_unit (Model240Units):
                    * Member of Model240Units IntEnum class
                    * Specifies the units to use for the data in this slot

        """

        self.slot_channel = channel
        self.slot_units = temp_unit


class Model240(GenericInstrument):
    """A class object representing the Lake Shore Model 240 channel modules"""

    vid_pid = [(0x1FB9, 0x0205)]

    def __init__(self,
                 serial_number=None,
                 com_port=None,
                 timeout=2.0,
                 **kwargs):
        # Call the parent init, then fill in values specific to the 240.
        GenericInstrument.__init__(self, serial_number, com_port, 115200, 8, 1, serial.PARITY_NONE, False,
                                   False, timeout, None, None, **kwargs)

    def get_identification(self):
        """Returns instrument's identification parameters.

            Returns:
                id (list):
                    List defining instrument's manufacturer, model, instrument serial, firmware version

        """
        id_parameter = self.query("*IDN?").split(",")
        id_list = {'manufacturer': id_parameter[0],
                   'model': id_parameter[1],
                   'serial number': id_parameter[2],
                   'firmware version': id_parameter[3]}

        return id_list

    def set_brightness(self, brightness_level):
        """Sets the brightness for the front panel display

            Args:
                brightness_level (Model240BrightnessLevel):
                    * Display brightness in percent

        """
        self.command(f"BRIGT {brightness_level}")

    def get_brightness(self):
        """Returns the brightness level of front panel display

            Return:
                brightness_level (Model240BrightnessLevel):
                    * Display brightness in percent

        """
        brightness_level = Model240BrightnessLevel(int(self.query("BRIGT?")))
        return brightness_level

    def get_celsius_reading(self, channel):
        """Returns the temperature value in Celsius of channel selected.

            Args:
                channel (int):
                    Specifies channel (1-8)

        """
        return self.query(f"CRDG? {channel}")

    def set_factory_defaults(self):
        """Sets all configuration values to factory defaults and resets the instrument"""
        self.command("DFLT 99")

    def get_kelvin_reading(self, channel):
        """Returns the temperature value in Kelvin of channel selected.

            Args:
                channel (int):
                    Specifies channel (1-8)

        """
        return float(self.query(f"KRDG? {channel}"))

    def get_fahrenheit_reading(self, channel):
        """Returns the temperature value in Farenheit of channel selected

            Args:
                channel (int):
                    Specifies channel (1-8)

        """
        return self.query(f"FRDG? {channel}")

    def get_sensor_reading(self, input_channel):
        """Returns the sensor reading in the sensor's units.

            Returns:
                reading (float):
                    * The raw sensor reading in the units of the connected sensor

        """
        return float(self.query(f"SRDG? {input_channel}"))

    def delete_curve(self, curve):
        """Deletes the user curve

            Args:
                curve (int):
                    * Specifies a user curve to delete

        """
        self.command(f"CRVDEL {curve}")

    def set_curve_header(self, input_channel, curve_header):
        """Configures the user curve header

            Args:
                input_channel (int):
                    * Specifies which input_channel curve to configure
                    * 1 - 8

                curve_header (CurveHeader):
                    * A CurveHeader class object containing the desired curve information

        """
        command_string = (f"CRVHDR {input_channel},{curve_header.curve_name},{curve_header.serial_number}," +
                            f"{curve_header.curve_data_format},{curve_header.temperature_limit},{curve_header.coefficient}")
        self.command(command_string)

    def get_curve_header(self, curve):
        """Returns parameters set on a particular user curve header

            Args:
                curve:
                    * Specifies a curve to retrieve

            Returns:
                header (CurveHeader):
                    * A CurveHeader class object containing the desired curve information

        """
        response = self.query(f"CRVHDR? {curve}")
        curve_header = response.split(",")
        header = Model240CurveHeader(str(curve_header[0]),
                                     str(curve_header[1]),
                                     Model240CurveFormat(int(curve_header[2])),
                                     float(curve_header[3]),
                                     Model240TemperatureCoefficient(int(curve_header[4])))
        return header

    def set_curve_data_point(self, channel, index, units, temp):
        """Configures a user curve point

            Args:
                channel (int):
                    Specifies which channel curve to configure (1-8)

                index (int):
                    Specifies the points index in the curve (1-200)

                units (float):
                    Specifies sensor units for this point to 6 digits

                temp (float):
                    Specifies the corresponding temperature in Kelvin for this point to 6 digits

        """
        self.command(f"CRVPT {channel},{index},{units},{temp}")

    def get_curve_data_point(self, channel, index):
        """Returns a standard or user curve data point

            Args:
                channel (int):
                    Specifies channel (1-8)

                index (int):
                    Specifies the points index in the curve (1–200)

        """
        return self.query(f"CRVPT? {channel},{index}")

    def set_filter(self, channel, length):
        """Sets the channel filter parameter

            Args:
                channel (int):
                    Specifies which channel to configure (1-8)

                length (int):
                    Specifies the number of 1 ms points to average for each update (1-100)

        """
        self.command(f"FILTER {channel},{length}")

    def get_filter(self, channel):
        """Returns the filter parameter

            Args:
                channel (int):
                    Specifies channel (1-8)

        """
        return self.query(f"FILTER? {channel}")

    def set_sensor_name(self, channel, name):
        """Names the sensor channel in specified channel.

            Args:
                channel (int):
                    Specifies which channel to configure (1-8)

                name (str):
                    Specifies the name to associate with the sensor channel

        """
        self.command(f"INNAME {channel},{name}")

    def get_sensor_name(self, channel):
        """Returns the sensor channel's name.

            Args:
                channel (int):
                    Specifies channel (1-8)

        """
        return str(self.query(f"INNAME? {channel}"))

    def set_input_parameter(self, channel, input_parameter):
        """Sets channel type parameters.

            Args:
                channel (int):
                    Specifies which channel to configure (1-8)

                input_parameter (InputParameter):
                    See InputParameter class

        """
        if input_parameter.auto_range_enable is None:
            input_range = 0
        else:
            input_range = input_parameter.input_range

        command_string = (f"INTYPE {channel},{input_parameter.sensor_type.value},{int(input_parameter.auto_range_enable)}," +
                            f"{input_range},{int(input_parameter.current_reversal_enable)},{input_parameter.temperature_unit.value}," +
                            f"{int(input_parameter.input_enable)}")

        self.command(command_string)

    def get_input_parameter(self, channel):
        """Returns channel type parameter details

            Args:
                channel (int):
                    Specifies channel (1-8)

        """
        response = self.query(f"INTYPE? {channel}")
        data = response.split(",")
        input_parameter = Model240InputParameter(Model240SensorTypes(int(data[0])),
                                                 bool(data[1]),
                                                 bool(data[3]),
                                                 Model240Units(int(data[4])),
                                                 bool(data[5]),
                                                 int(data[2]))
        return input_parameter

    def set_modname(self, name):
        """Names module.

            Args:
                name (str):
                    Specifies the name or description to help identify the module

        """
        self.command(f"MODNAME {name}")

    def get_modname(self):
        """Returns module name

            Returns:
                modname (str):
                    Specifies name of module

        """
        return self.query("MODNAME?")

    def set_profibus_slot_count(self, count):
        """Configures the number of PROFIBUS slots for the instrument to present to the bus as a modular station

            Args:
                count (int):
                    Specifies the number of PROFIBUS slots (1-8)

        """
        self.command(f"PROFINUM {count}")

    def get_profibus_slot_count(self):
        """Returns the number of PROFIBUS slots for the instrument present to the bus as a modular station

            Returns:
                slot_count (str):
                    Specifies PROFIBUS slot count

        """
        return self.query("PROFINUM?")

    def set_profibus_address(self, address):
        """Configures the PROFIBUS address for the module. An address of 126 indicates that it is not configured and it
        then can be set by a PROFIBUS master.

            Args:
                address (str):
                    Specifies the PROFIBUS address (1-126)

        """
        self.command(f"ADDR {address}")

    def get_profibus_address(self):
        """Returns the PROFIBUS address for the module.

            Returns:
                address (str):
                    Specifies PROFIBUS address of module

        """
        return int(self.query("ADDR?"))

    def set_profibus_slot_configuration(self, slot, profislot_config):
        """Configures what data to present on the given PROFIBUS slot. Note that the correct
            number of slots must be configured with the PROFINUM command, or the slot may be ignored.

            Args:
                slot (int):
                    Specifies the slot to be configured

                profislot_config (Model240ProfiSlot):
                    A Model240ProfiSlot class object containing the desired PROFIBUS slot configuration information

        """
        self.command(f"PROFISLOT {slot},{profislot_config.slot_channel},{profislot_config.slot_units}")

    def get_profibus_slot_configuration(self, slot_num):
        """Returns the slot configuration of the slot number.

            Returns:
                slot_config (Model240ProfiSlot):
                    See Model240ProfiSlot class

        """
        response = self.query(f"PROFISLOT? {slot_num}").split(",")
        slot_configuration = Model240ProfiSlot(int(response[0]), Model240Units(int(response[1])))
        return slot_configuration

    def get_profibus_connection_status(self):
        """Returns the connection status of PROFIBUS.

            Returns:
                status (str):
                    Specifies connection status of PROFIBUS

        """
        return self.query("PROFISTAT?")

    def get_channel_reading_status(self, channel):
        """The integer returned represents the sum of the bit weighting of the channel status flag
            bits. A “000” response indicates a valid reading is present.

            Args:
                channel (int):
                    Specifies which channel to query (1-8)

            Returns:
                bit_status (dict):
                    Dictionary containing the current status indicator

        """
        bit_status = {}
        bit_names = ["invalid reading",
                     "",
                     "",
                     "",
                     "temp under range",
                     "temp over range",
                     "sensor units over range",
                     "sensor units under range"
                     ]

        status_indicator = int(self.query(f"RDGST? {channel}"))

        for count, bit_name in enumerate(bit_names):
            mask = 0b1 << count
            bit_status[bit_name] = bool(mask & status_indicator)

        return bit_status

    def get_sensor_units_channel_reading(self, channel):
        """Returns the sensor units value of channel being queried

            Args:
                channel (int):
                    Specifies which channel to query (1-8)

        """
        return self.query(f"SRDG? {channel}")


__all__ = ['Model240', 'Model240Units', 'Model240CurveFormat', 'Model240Coefficients', 'Model240SensorTypes',
           'Model240BrightnessLevel', 'Model240TemperatureCoefficient', 'Model240InputRange', 'Model240CurveHeader',
           'Model240InputParameter', 'Model240ProfiSlot']



"""Implements functionality unique to the M81 Measure Modules."""

from datetime import datetime
from warnings import warn
from lakeshore.xip_instrument import RegisterBase
from lakeshore.ssm_base_module import SSMSystemModuleQuestionableRegister, BaseModule


class SSMSystemMeasureModuleOperationRegister(RegisterBase):
    """Class object representing the operation status register of a measure module"""

    bit_names = [
        "overload",
        "settling",
        "unlocked"
    ]

    def __init__(
            self,
            overload,
            settling,
            unlocked):
        self.overload = overload
        self.settling = settling
        self.unlocked = unlocked

# pylint: disable=R0904
class MeasureModule(BaseModule):
    """Class for interaction with a specific measure channel of the M81 instrument"""

    def get_name(self):
        """Returns the user-settable name of the module"""

        return self.device.query(f'SENSe{self.module_number}:NAME?').strip('\"')

    def set_name(self, new_name):
        """Set the name of the module"""

        self.device.command(f'SENSe{self.module_number}:NAME "{new_name}"')

    def get_notes(self):
        """Returns the user-settable notes of the module"""

        return self.device.query(f'SENSe{self.module_number}:NOTes?').strip('\"')

    def set_notes(self, new_note):
        """Set the notes of the module"""

        self.device.command(f'SENSe{self.module_number}:NOTes "{new_note}"')

    def get_model(self):
        """Returns the model of the module (i.e. VM-10)"""

        return self.device.query(f'SENSe{self.module_number}:MODel?').strip('\"')

    def get_serial(self):
        """Returns the serial number of the module (i.e. LSA1234)"""

        return self.device.query(f'SENSe{self.module_number}:SERial?').strip('\"')

    def get_hw_version(self):
        """Returns the hardware version of the module"""

        return int(self.device.query(f'SENSe{self.module_number}:HWVersion?'))

    def get_self_cal_status(self):
        """Returns the status of the last self calibration of the module"""

        return self.device.query(f'SENSe{self.module_number}:SCALibration:STATus?')

    def run_self_cal(self):
        """Run a self calibration for the module"""

        self.device.command(f'SENSe{self.module_number}:SCALibration:RUN')

    def reset_self_cal(self):
        """Restore factory self calibration for the module"""

        self.device.command(f'SENSe{self.module_number}:SCALibration:RESet')

    def get_averaging_time(self):
        """Returns the averaging time of the module in Power Line Cycles. Not relevant in lock-in mode."""

        return float(self.device.query(f'SENSe{self.module_number}:NPLCycles?'))

    def set_averaging_time(self, nplc):
        """Sets the averaging time of the module. Not relevant in lock-in mode.

            Args:
                nplc (float):
                    The new number of power line cycles to average
        """

        self.device.command(f'SENSe{self.module_number}:NPLCycles {float(nplc)}')

    def get_mode(self):
        """Returns the measurement mode of the module. 'DC', 'AC', or 'LIA'."""

        return self.device.query(f'SENSe{self.module_number}:MODE?')

    def set_mode(self, mode):
        """Sets the measurement mode of the module

            Args:
                mode (str):
                    The new measurement mode ('DC', 'AC', or 'LIA')
        """

        self.device.command(f'SENSe{self.module_number}:MODE {mode}')

    def get_coupling(self):
        """Return input coupling of the module. 'AC' or 'DC'."""

        return self.device.query(f'SENSe{self.module_number}:COUPling?')

    def set_coupling(self, coupling):
        """Sets the input coupling of the module

            Args:
                coupling (str):
                    The new input coupling ('AC' or 'DC')
        """

        self.device.command(f'SENSe{self.module_number}:COUPling {coupling}')

    def use_ac_coupling(self):
        """Sets the input coupling of the module to 'AC'"""

        self.set_coupling('AC')

    def use_dc_coupling(self):
        """Sets the input coupling of the module to 'DC'"""

        self.set_coupling('DC')

    def get_input_configuration(self):
        """Returns the input configuration of the module. 'AB', 'A', or 'GROUND'."""
        return self.device.query(f'SENSe{self.module_number}:CONFiguration?')

    def set_input_configuration(self, input_configuration):
        """Sets the input configuration of the module

            Args:
                input_configuration (str):
                    The new input configuration ('AB', 'A', or 'GROUND')
        """

        self.device.command(f'SENSe{self.module_number}:CONFiguration {input_configuration}')

    def enable_bias_voltage(self):
        """Enables the bias voltage applied to the amplifier"""

        self.device.command(f'SENSe{self.module_number}:BIAS:STATe 1')

    def disable_bias_voltage(self):
        """Disables the bias voltage applied to the amplifier"""

        self.device.command(f'SENSe{self.module_number}:BIAS:STATe 0')

    def get_bias_voltage_enabled(self):
        """Return whether the bias voltage is enabled"""

        return bool(int(self.device.query(f'SENSe{self.module_number}:BIAS:STATe?')))

    def get_bias_voltage(self):
        """Return the bias voltage applied on the amplifier input in Volts"""

        return float(self.device.query(f'SENSe{self.module_number}:BIAS:VOLTage:DC?'))

    def set_bias_voltage(self, bias_voltage):
        """Sets the bias voltage applied on the amplifier input

            Args:
                bias_voltage (float):
                    The new bias voltage in Volts
        """

        self.device.command(f'SENSe{self.module_number}:BIAS:VOLTage:DC {str(bias_voltage)}')

    def get_filter_state(self):
        """Returns whether the hardware filter is engaged"""

        return bool(int(self.device.query(f'SENSe{self.module_number}:FILTer:STATe?')))

    def get_lowpass_corner_frequency(self):
        """Returns the low pass filter cuttoff frequency. 'NONE', 'F10', 'F30', 'F100', 'F300', 'F1000', 'F3000', or 'F10000'."""

        return self.device.query(f'SENSe{self.module_number}:FILTer:LPASs:FREQuency?')

    def get_lowpass_rolloff(self):
        """Returns the low pass filter roll-off. 'R6' or 'R12'."""

        return self.device.query(f'SENSe{self.module_number}:FILTer:LPASs:ATTenuation?')

    def get_highpass_corner_frequency(self):
        """Returns the high pass filter cuttoff frequency. 'NONE', 'F10', 'F30', 'F100', 'F300', 'F1000', or 'F3000'."""

        return self.device.query(f'SENSe{self.module_number}:FILTer:HPASs:FREQuency?')

    def get_highpass_rolloff(self):
        """Returns the high pass filter roll-off. 'R6' or 'R12'."""

        return self.device.query(f'SENSe{self.module_number}:FILTer:HPASs:ATTenuation?')

    def get_gain_allocation_strategy(self):
        """Returns the gain allocation strategy used for the hardware filter. 'NOISE', or 'RESERVE'."""

        return self.device.query(f'SENSe{self.module_number}:FILTer:OPTimization?')

    def set_gain_allocation_strategy(self, optimization_type):
        """Sets the gain allocation strategy used for the hardware filter

            Args:
                optimization_type (str):
                    The new optimization type ('NOISE', or 'RESERVE')
        """

        self.device.command(f'SENSe{self.module_number}:FILTer:OPTimization {optimization_type}')

    def configure_input_lowpass_filter(self, corner_frequency, rolloff='R12'):
        """Configure the input low pass filter

            Args:
                corner_frequency (str):
                    The low pass corner frequency ('NONE', 'F10', 'F30', 'F100', 'F300', 'F1000', 'F3000', or 'F10000'). F10 = 10 Hz, etc.

                rolloff (str):
                    The low pass roll-off ('R6' or 'R12'). R6 = 6 dB/Octave, R12 = 12 dB/Octave.
        """

        self.device.command(f'SENSe{self.module_number}:FILTer:LPASs:FREQuency {corner_frequency}')
        self.device.command(f'SENSe{self.module_number}:FILTer:LPASs:ATTenuation {rolloff}')
        self.device.command(f'SENSe{self.module_number}:FILTer:STATe 1')

    def configure_input_highpass_filter(self, corner_frequency, rolloff='R12'):
        """Configure the input high pass filter

            Args:
                corner_frequency (str):
                    The high pass corner frequency ('NONE', 'F10', 'F30', 'F100', 'F300', 'F1000', or 'F3000'). F10 = 10 Hz, etc.

                rolloff (str):
                    The high pass roll-off ('R6' or 'R12'). R6 = 6 dB/Octave, R12 = 12 dB/Octave.
        """

        self.device.command(f'SENSe{self.module_number}:FILTer:HPASs:FREQuency {corner_frequency}')
        self.device.command(f'SENSe{self.module_number}:FILTer:HPASs:ATTenuation {rolloff}')
        self.device.command(f'SENSe{self.module_number}:FILTer:STATe 1')

    def disable_input_filters(self):
        """Disables the hardware filters"""

        self.device.command(f'SENSe{self.module_number}:FILTer:STATe 0')

    def get_current_range(self):
        """Returns the current range in Amps"""

        return float(self.device.query(f'SENSe{self.module_number}:CURRent:RANGe?'))

    def get_i_range(self):
        """
        Returns the current range in Amps

        .. deprecated:: 1.5.4
           Use get_current_range instead
        """

        warn('The get_i_range method is deprecated. Use get_current_range instead.', DeprecationWarning)
        return self.get_current_range()

    def get_current_autorange_status(self):
        """Returns whether autoranging is enabled for the module"""

        return bool(int(self.device.query(f'SENSe{self.module_number}:CURRent:RANGe:AUTO?')))

    def get_i_autorange_status(self):
        """
        Returns whether autoranging is enabled for the module

        .. deprecated:: 1.5.4
           Use get_current_autorange_status instead
        """

        warn('The get_i_autorange_status method is deprecated. Use get_current_autorange_status instead.', DeprecationWarning)
        return self.get_current_autorange_status()

    def configure_current_range(self, autorange, max_level=None):
        """Configure current ranging for the module

            Args:
                autorange (bool):
                    True to enable real time range decisions by the module. False for manual ranging.

                max_level (float):
                    The largest current that needs to be measured by the module in Amps.
        """

        if autorange:
            if max_level is not None:
                raise ValueError('If autorange is selected, a manual range cannot be specified.')

            self.device.command(f'SENSe{self.module_number}:CURRent:RANGe:AUTO 1')
        else:
            if max_level is not None:
                self.device.command(f'SENSe{self.module_number}:CURRent:RANGe {str(max_level)}')

    def configure_i_range(self, autorange, max_level=None):
        """
        Configure current ranging for the module

        .. deprecated:: 1.5.4
           Use configure_current_range instead

            Args:
                autorange (bool):
                    True to enable real time range decisions by the module. False for manual ranging.

                max_level (float):
                    The largest current that needs to be measured by the module in Amps.
        """

        warn('The configure_i_range method is deprecated. Use configure_current_range instead.', DeprecationWarning)
        self.configure_current_range(autorange, max_level)

    def get_voltage_range(self):
        """Returns the voltage range in Volts"""

        return float(self.device.query(f'SENSe{self.module_number}:VOLTage:RANGe?'))

    def get_voltage_autorange_status(self):
        """Returns whether autoranging is enabled for the module"""

        return bool(int(self.device.query(f'SENSe{self.module_number}:VOLTage:RANGe:AUTO?')))

    def configure_voltage_range(self, autorange, max_level):
        """Configure voltage ranging for the module

            Args:
                autorange (bool):
                    True to enable real time range decisions by the module. False for manual ranging.

                max_level (float):
                    The largest voltage that needs to be measured by the module in Volts.
        """

        if autorange:
            if max_level is not None:
                raise ValueError('If autorange is selected, a manual range cannot be specified.')

            self.device.command(f'SENSe{self.module_number}:VOLTage:RANGe:AUTO 1')
        else:
            if max_level is not None:
                self.device.command(f'SENSe{self.module_number}:VOLTage:RANGe {str(max_level)}')

    def get_reference_source(self):
        """Returns the lock-in reference source. 'S1', 'S2', 'S3', 'RIN'."""

        return self.device.query(f'SENSe{self.module_number}:LIA:RSOurce?')

    def set_reference_source(self, reference_source):
        """Sets the lock-in reference source

            Args:
                reference_source (str):
                    The new reference source ('S1', 'S2', 'S3', 'RIN')
        """

        self.device.command(f'SENSe{self.module_number}:LIA:RSOurce {reference_source}')

    def get_reference_harmonic(self):
        """Returns the lock-in reference harmonic"""

        return int(self.device.query(f'SENSe{self.module_number}:LIA:DHARmonic?'))

    def set_reference_harmonic(self, harmonic):
        """Sets the lock-in reference harmonic

            Args:
                harmonic (int):
                    The new reference harmonic. 1 is the fundamental frequency, 2 is twice the fundamental frequency, etc.
        """

        self.device.command(f'SENSe{self.module_number}:LIA:DHARmonic {str(harmonic)}')

    def get_reference_phase_shift(self):
        """Returns the lock-in reference phase shift in degrees"""

        return float(self.device.query(f'SENSe{self.module_number}:LIA:DPHase?'))

    def set_reference_phase_shift(self, phase_shift):
        """Sets the lock-in reference phase shift

            Args:
                phase_shift (float):
                    The new reference phase shift in degrees
        """
        self.device.command(f'SENSe{self.module_number}:LIA:DPHase {str(phase_shift)}')

    def auto_phase(self):
        """Executes a one time adjustment of the reference phase shift such that the present phase measurement is zero."""

        self.device.command(f'SENSe{self.module_number}:LIA:DPHase:AUTO')

    def get_lock_in_time_constant(self):
        """Returns the lock-in time constant in seconds"""

        return float(self.device.query(f'SENSe{self.module_number}:LIA:TIMEconstant?'))

    def set_lock_in_time_constant(self, time_constant):
        """Sets the lock-in time constant

            Args:
                time_constant (float):
                    The new time constant in seconds
        """
        self.device.command(f'SENSe{self.module_number}:LIA:TIMEconstant {str(time_constant)}')

    def get_lock_in_settle_time(self, settle_percent=0.01):
        """Returns the lock-in settle time in seconds

            Args:
                settle_percent (float)
                    The desired percent signal has settled to in percent
                    A value of `0.1` is interpreted as 0.1 %
        """
        return float(self.device.query(f'SENSe{self.module_number}:LIA:STIMe? {str(settle_percent)}'))

    def get_lock_in_equivalent_noise_bandwidth(self):
        """Returns the equivalent noise bandwidth (ENBW) in Hz"""
        return float(self.device.query(f'SENSe{self.module_number}:LIA:ENBW?'))

    def get_lock_in_rolloff(self):
        """Returns the lock-in PSD output filter roll-off for the present module. 'R6', 'R12', 'R18' or 'R24'."""

        return self.device.query(f'SENSe{self.module_number}:LIA:ROLLoff?')

    def set_lock_in_rolloff(self, rolloff):
        """Sets the lock-in PSD output filter roll-off

            Args:
                rolloff (str):
                    The new PSD output filter roll-off ('R6', 'R12', 'R18' or 'R24')
        """

        self.device.command(f'SENSe{self.module_number}:LIA:ROLLoff {rolloff}')

    def get_lock_in_iir_state(self):
        """Returns the state of the lock-in PSD output IIR filter"""

        return bool(int(self.device.query(f'SENSe{self.module_number}:LIA:IIR:STATe?')))

    def set_lock_in_iir_state(self, state):
        """Sets the state of the lock-in PSD output IIR filter

            Args:
                state (bool):
                    The new state of the PSD output IIR filter
        """

        self.device.command(f'SENSe{self.module_number}:LIA:IIR:STATe {str(int(state))}')

    def enable_lock_in_iir(self):
        """Sets the state of the lock-in PSD output IIR filter to True."""

        self.set_lock_in_iir_state(True)

    def disable_lock_in_iir(self):
        """Sets the state of the lock-in PSD output IIR filter to False."""

        self.set_lock_in_iir_state(False)

    def get_lock_in_fir_state(self):
        """Returns the state of the lock-in PSD output FIR filter"""

        return bool(int(self.device.query(f'SENSe{self.module_number}:LIA:FIR:STATe?')))

    def set_lock_in_fir_state(self, state):
        """Sets the state of the lock-in PSD output FIR filter

            Args:
                state (bool):
                    The new state of the PSD output FIR filter
        """

        self.device.command(f'SENSe{self.module_number}:LIA:FIR:STATe {str(int(state))}')

    def enable_lock_in_fir(self):
        """Sets the state of the lock-in PSD output FIR filter to True."""

        self.set_lock_in_fir_state(True)

    def disable_lock_in_fir(self):
        """Sets the state of the lock-in PSD output FIR filter to False."""

        self.set_lock_in_fir_state(False)

    def get_lock_in_fir_cycles(self):
        """Returns the number of FIR cycles"""

        return int(self.device.query(f'SENSe{self.module_number}:LIA:FIR:CYCLes?'))

    def set_lock_in_fir_cycles(self, cycles):
        """Sets the number of FIR cycles

            Args:
                cycles (int):
                    The desired number of FIR cycles, between 1 and 100
        """

        self.device.command(f'SENSe{self.module_number}:LIA:FIR:CYCLes {str(int(cycles))}')

    def setup_dc_measurement(self, nplc=1):
        """Setup the module for DC measurement

            Args:
                nplc (float):
                    The new number of power line cycles to average
        """

        self.set_mode('DC')
        self.set_averaging_time(nplc)

    def setup_ac_measurement(self, nplc=1):
        """Setup the module for DC measurement

            Args:
                nplc (float):
                    The new number of power line cycles to average
        """

        self.set_mode('AC')
        self.set_averaging_time(nplc)

    def setup_lock_in_measurement(self,
                                  reference_source,
                                  time_constant,
                                  rolloff='R24',
                                  reference_phase_shift=0.0,
                                  reference_harmonic=1,
                                  use_fir=True):
        """Setup the module for lock-in measurement

            Args:
                reference_source (str):
                    Lock-in reference source ('S1', 'S2', 'S3', 'RIN')

                time_constant (float):
                    Time constant in seconds

                rolloff (str):
                    Lock-in PSD output filter roll-off ('R6', 'R12', 'R18' or 'R12')

                reference_phase_shift (float):
                    Lock-in reference phase shift in degrees

                reference_harmonic (int):
                    Lock-in reference harmonic. 1 is the fundamental frequency, 2 is twice the fundamental frequency, etc.

                use_fir (bool):
                    Enable or disable the PSD output FIR filter
        """

        self.set_mode('LIA')
        self.set_reference_source(reference_source)
        self.set_lock_in_time_constant(time_constant)
        self.set_lock_in_rolloff(rolloff)
        self.set_reference_phase_shift(reference_phase_shift)
        self.set_reference_harmonic(reference_harmonic)
        self.set_lock_in_fir_state(use_fir)

    def zero_relative_baseline(self):
        """Sets the present measurement as the baseline value for calculating relative readings"""

        self.device.command(f'SENSe{self.module_number}:RELative:ZERO')

    def set_relative_baseline(self, baseline):
        """Sets the relative baseline"""

        self.device.command(f'SENSe{self.module_number}:RELative:BASEline {str(baseline)}')

    def get_relative_baseline(self):
        """Returns the relative baseline"""

        return float(self.device.query(f'SENSe{self.module_number}:RELative:BASEline?'))

    def get_multiple(self, *data_sources):
        """This function is deprecated. Use fetch_multiple() instead."""

        return self.fetch_multiple(*data_sources)

    def get_dc(self):
        """Returns the DC measurement in module units"""

        return float(self.device.query(f'READ:SENSe{self.module_number}:DC?'))

    def get_dc_relative(self):
        """Returns the relative DC measurement in module units"""

        return float(self.device.query(f'READ:SENSe{self.module_number}:DC:RELative?'))

    def get_dc_minimum(self):
        """Returns the minimum DC indication in module units"""

        return float(self.device.query(f'STATistic:MINimum:SENSe{self.module_number}:DC?'))

    def get_dc_maximum(self):
        """Returns the maximum DC indication in module units"""

        return float(self.device.query(f'STATistic:MAXimum:SENSe{self.module_number}:DC?'))

    def get_rms(self):
        """Returns the RMS measurement in module units"""

        return float(self.device.query(f'READ:SENSe{self.module_number}:RMS?'))

    def get_rms_relative(self):
        """Returns the relative RMS measurement in module units"""

        return float(self.device.query(f'READ:SENSe{self.module_number}:RMS:RELative?'))

    def get_rms_minimum(self):
        """Returns the minimum RMS indication in module units"""

        return float(self.device.query(f'STATistic:MINimum:SENSe{self.module_number}:RMS?'))

    def get_rms_maximum(self):
        """Returns the maximum RMS indication in module units"""

        return float(self.device.query(f'STATistic:MAXimum:SENSe{self.module_number}:RMS?'))

    def get_peak_to_peak(self):
        """Returns the peak to peak measurement in module units"""

        return float(self.device.query(f'READ:SENSe{self.module_number}:PTPeak?'))

    def get_peak_to_peak_minimum(self):
        """Returns the minimum peak to peak indication in module units"""

        return float(self.device.query(f'STATistic:MINimum:SENSe{self.module_number}:PTPeak?'))

    def get_peak_to_peak_maximum(self):
        """Returns the maximum peak to peak indication in module units"""

        return float(self.device.query(f'STATistic:MAXimum:SENSe{self.module_number}:PTPeak?'))

    def get_positive_peak(self):
        """Returns the positive peak measurement in module units"""

        return float(self.device.query(f'READ:SENSe{self.module_number}:PPEak?'))

    def get_positive_peak_minimum(self):
        """Returns the minimum positive peak indication in module units"""

        return float(self.device.query(f'STATistic:MINimum:SENSe{self.module_number}:PPEak?'))

    def get_positive_peak_maximum(self):
        """Returns the maximum positive peak indication in module units"""

        return float(self.device.query(f'STATistic:MAXimum:SENSe{self.module_number}:PPEak?'))

    def get_negative_peak(self):
        """Returns the negative peak measurement in module units"""

        return float(self.device.query(f'READ:SENSe{self.module_number}:NPEak?'))

    def get_negative_peak_minimum(self):
        """Returns the minimum negative peak indication in module units"""

        return float(self.device.query(f'STATistic:MINimum:SENSe{self.module_number}:NPEak?'))

    def get_negative_peak_maximum(self):
        """Returns the maximum negative peak indication in module units"""

        return float(self.device.query(f'STATistic:MAXimum:SENSe{self.module_number}:NPEak?'))

    def get_lock_in_x(self):
        """Returns the present X measurement from the lock-in"""

        return float(self.device.query(f'FETCh:SENSe{self.module_number}:LIA:X?'))

    def get_lock_in_x_minimum(self):
        """Returns the minimum X indication from the lock in"""

        return float(self.device.query(f'STATistic:MINimum:SENSe{self.module_number}:LIA:X?'))

    def get_lock_in_x_maximum(self):
        """Returns the maximum X indication from the lock in"""

        return float(self.device.query(f'STATistic:MAXimum:SENSe{self.module_number}:LIA:X?'))

    def get_lock_in_y(self):
        """Returns the present Y measurement from the lock-in"""

        return float(self.device.query(f'FETCh:SENSe{self.module_number}:LIA:Y?'))

    def get_lock_in_y_minimum(self):
        """Returns the minimum Y indication from the lock in"""

        return float(self.device.query(f'STATistic:MINimum:SENSe{self.module_number}:LIA:Y?'))

    def get_lock_in_y_maximum(self):
        """Returns the maximum Y indication from the lock in"""

        return float(self.device.query(f'STATistic:MAXimum:SENSe{self.module_number}:LIA:Y?'))

    def get_lock_in_r(self):
        """Returns the present magnitude measurement from the lock-in"""

        return float(self.device.query(f'FETCh:SENSe{self.module_number}:LIA:R?'))

    def get_lock_in_r_minimum(self):
        """Returns the minimum magnitude indication from the lock in"""

        return float(self.device.query(f'STATistic:MINimum:SENSe{self.module_number}:LIA:R?'))

    def get_lock_in_r_maximum(self):
        """Returns the maximum magnitude indication from the lock in"""

        return float(self.device.query(f'STATistic:MAXimum:SENSe{self.module_number}:LIA:R?'))

    def get_lock_in_theta(self):
        """Returns the present angle measurement from the lock-in"""

        return float(self.device.query(f'FETCh:SENSe{self.module_number}:LIA:THETa?'))

    def get_lock_in_theta_minimum(self):
        """Returns the minimum angle indication from the lock in"""

        return float(self.device.query(f'STATistic:MINimum:SENSe{self.module_number}:LIA:THETa?'))

    def get_lock_in_theta_maximum(self):
        """Returns the maximum angle indication from the lock in"""

        return float(self.device.query(f'STATistic:MAXimum:SENSe{self.module_number}:LIA:THETa?'))

    def get_lock_in_frequency(self):
        """Returns the present detected frequency from the Phase Locked Loop (PLL)"""

        return float(self.device.query(f'FETCh:SENSe{self.module_number}:LIA:FREQuency?'))

    def get_pll_lock_status(self):
        """Returns the present lock status of the PLL. True if locked, False if unlocked."""

        return bool(int(self.device.query(f'FETCh:SENSe{self.module_number}:LIA:LOCK?')))

    def get_present_questionable_status(self):
        """Returns the names of the questionable status register bits and their values"""

        response = self.device.query(f'STATus:QUEStionable:SENSe{self.module_number}:CONDition?', check_errors=False)
        status_register = SSMSystemModuleQuestionableRegister.from_integer(response)

        return status_register

    def get_questionable_events(self):
        """Returns the names of questionable event status register bits that are currently high.
        The event register is latching and values are reset when queried."""

        response = self.device.query(f'STATus:QUEStionable:SENSe{self.module_number}:EVENt?', check_errors=False)
        status_register = SSMSystemModuleQuestionableRegister.from_integer(response)

        return status_register

    def get_questionable_event_enable_mask(self):
        """Returns the names of the questionable event enable register bits and their values.
        These values determine which questionable bits propagate to the questionable event register."""

        response = self.device.query(f'STATus:QUEStionable:SENSe{self.module_number}:ENABle?', check_errors=False)
        status_register = SSMSystemModuleQuestionableRegister.from_integer(response)

        return status_register

    def set_questionable_event_enable_mask(self, register_mask):
        """Configures the values of the questionable event enable register bits.
        These values determine which questionable bits propagate to the questionable event register.

            Args:
                register_mask ([Instrument]QuestionableRegister):
                    An instrument specific QuestionableRegister class object with all bits configured true or false.
        """

        integer_representation = register_mask.to_integer()
        self.device.command(f'STATus:QUEStionable:SENSe{self.module_number}:ENABle {integer_representation}', check_errors=False)

    def get_present_operation_status(self):
        """Returns the names of the operation status register bits and their values"""

        response = self.device.query(f'STATus:OPERation:SENSe{self.module_number}:CONDition?', check_errors=False)
        status_register = SSMSystemMeasureModuleOperationRegister.from_integer(response)

        return status_register

    def get_operation_events(self):
        """Returns the names of operation event status register bits that are currently high.
        The event register is latching and values are reset when queried."""

        response = self.device.query(f'STATus:OPERation:SENSe{self.module_number}:EVENt?', check_errors=False)
        status_register = SSMSystemMeasureModuleOperationRegister.from_integer(response)

        return status_register

    def get_operation_event_enable_mask(self):
        """Returns the names of the operation event enable register bits and their values.
        These values determine which operation bits propagate to the operation event register."""

        response = self.device.query(f'STATus:OPERation:SENSe{self.module_number}:ENABle?', check_errors=False)
        status_register = SSMSystemMeasureModuleOperationRegister.from_integer(response)

        return status_register

    def set_operation_event_enable_mask(self, register_mask):
        """Configures the values of the operation event enable register bits.
        These values determine which operation bits propagate to the operation event register.

            Args:
                register_mask ([Instrument]OperationRegister):
                    An instrument specific OperationRegister class object with all bits configured true or false.
        """

        integer_representation = register_mask.to_integer()
        self.device.command(f'STATus:OPERation:SENSe{self.module_number}:ENABle {integer_representation}', check_errors=False)

    def get_identify_state(self):
        """Returns the identification state for the given pod."""
        response = bool(int(self.device.query(f'SENSe{self.module_number}:IDENtify?', check_errors=False)))
        return response

    def set_identify_state(self, state):
        """Sets the identification state for the given pod.

            Args:
                state (bool):
                    The desired state for the LED, 1 for identify, 0 for normal state
        """
        self.device.command(f'SENSe{self.module_number}:IDENtify {int(state)}', check_errors=False)

    def get_dark_mode_state(self):
        """Returns the dark mode state for the given pod"""
        response = self.device.query(f'SENSe{self.module_number}:DMODe?', cherk_errors=False)
        return response

    def set_dark_mode_state(self, state):
        """Configures the dark mode state for the given pod.

            Args:
                state (bool):
                    The desired operation for the LED, 1 for normal mode, 0 for dark mode
        """
        self.device.command(f'SENSe{self.module_number}:DMODe {state}', check_errors=False)

    def get_frequency_range_threshold(self):
        """Returns the frequency range threshold for the module. Frequency range threshold normalized to the -3 db
           point. For example, a value of 0.1 means 10 % of the -3 db point.
        """

        return float(self.device.query(f'SENSe{self.module_number}:FRTHreshold?'))

    def set_frequency_range_threshold(self, threshold):
        """Sets the frequency range threshold for the specified module.
        When the modules range is set to Auto, a range such that the frequency of the signal does not exceed the given
        percentage of the bandwidth of the range will be chosen.

        Args:
            threshold (float):
                Frequency range threshold normalized to the -3 db point with a valid range of 0.0 to 1.0.
                For example, a value of 0.1 means 10 % of the -3 db point.
        """

        self.device.command(f'SENSe{self.module_number}:FRTHreshold {float(threshold)}')

    def get_digital_high_pass_filter_state(self):
        """Returns the state of the digital high pass filter for lock-in mode."""
        return bool(int(self.device.query(f'SENSe{self.module_number}:DIGital:FILTer:HPASs:STATe?')))

    def set_digital_high_pass_filter_state(self, state):
        """Sets the state of the digital high pass filter for lock-in mode.

            Args:
                state (bool):
                    The desired state fot he digital lock-in high pass filter. 1 for enabled, 0 for disabled.
        """
        self.device.command(f'SENSe{self.module_number}:DIGital:FILTer:HPASs:STATE {int(state)}')

    def get_resistance(self):
        """Returns the present resistance measurement in Ohms.  A valid source must be configured."""

        return float(self.device.query(f'CALCulate:SENSe{self.module_number}:RESistance?'))

    def set_resistance_source(self, source):
        """Configures the resistance feature to use a specified source module to calculate resistance.

            Args:
                source (str):
                    The channel used for calculating resistance ('S1', 'S2', or 'S3').
        """

        self.device.command(f'CALCulate:SENSe{self.module_number}:RESistance:SOURce {source}')

    def get_resistance_source(self):
        """Returns the present source module being used to calculate resistance."""

        return self.device.query(f'CALCulate:SENSe{self.module_number}:RESistance:SOURce?')

    def reset_settings(self):
        """Resets the settings for the specified module to their power on defaults.
        """

        self.device.command(f'SENSe{self.module_number}:PRESet')

    def unload(self):
        """Unloads the specified module.
        """

        self.device.command(f'SENSe{self.module_number}:UNLoad')

    def get_load_state(self):
        """Returns the loaded state for the specified module.
        """

        response = bool(int(self.device.query(f'SENSe{self.module_number}:LOAD?')))
        return response

    def fetch_multiple(self, *data_sources):
        """Gets a list of the latest values corresponding to the input data sources for this module, and returns them
        as soon as possible.

            Args:
                data_sources (SSMSystemDataSourceMnemonic or str): Variable length list of data sources.

            Returns:
                Tuple of values corresponding to the given data sources for this module
        """

        elements = [(data_source, self.module_number) for data_source in data_sources]
        return self.device.fetch_multiple(*elements)

    def read_multiple(self, *data_sources):
        """Initiates measurement of new values corresponding to the input data sources for this module, and returns them
        after the measurement is complete.

            Args:
                data_sources (SSMSystemReadDataSourceMnemonic or str): Variable length list of data sources.

            Returns:
                Tuple of values corresponding to the given data sources for this module
        """

        elements = [(data_source, self.module_number) for data_source in data_sources]
        return self.device.read_multiple(*elements)

    def get_self_cal_datetime(self):
        """Returns the self calibration date and time for the specified module.
        """

        response = self.device.query(f'SENSe{self.module_number}:SCALibration:DATE?').split(',')
        return datetime(int(response[0]), int(response[1]), int(response[2]), int(response[3]),int(response[4]), int(response[5]))

    def get_self_cal_temperature(self):
        """Returns the self calibration temperature for the specified module.
        """

        return float(self.device.query(f'SENSe{self.module_number}:SCALibration:TEMP?'))
        


"""This module implements a parent class that contains all functionality shared by Lake Shore XIP instruments."""

import re

import serial

from .generic_instrument import GenericInstrument, InstrumentException, RegisterBase


class StatusByteRegister(RegisterBase):
    """Class object representing the status byte register"""

    bit_names = [
        "",
        "",
        "error_available",
        "questionable_summary",
        "message_available_summary",
        "event_status_summary",
        "master_summary",
        "operation_summary"
    ]

    def __init__(self,
                 error_available,
                 questionable_summary,
                 message_available_summary,
                 event_status_summary,
                 master_summary,
                 operation_summary):
        self.error_available = error_available
        self.questionable_summary = questionable_summary
        self.message_available_summary = message_available_summary
        self.event_status_summary = event_status_summary
        self.master_summary = master_summary
        self.operation_summary = operation_summary


class StandardEventRegister(RegisterBase):
    """Class object representing the standard event register"""

    bit_names = [
        "operation_complete",
        "query_error",
        "device_specific_error",
        "execution_error",
        "command_error",
        "",
        "power_on"
    ]

    def __init__(self,
                 operation_complete,
                 query_error,
                 device_specific_error,
                 execution_error,
                 command_error,
                 power_on):
        self.operation_complete = operation_complete
        self.query_error = query_error
        self.device_specific_error = device_specific_error
        self.execution_error = execution_error
        self.command_error = command_error
        self.power_on = power_on


class XIPInstrumentException(Exception):
    """Names a new type of exception specific to XIP instrument connectivity."""


class XIPInstrument(GenericInstrument):
    """Parent class that implements functionality shared by all XIP instruments"""

    def __init__(self,
                 serial_number,
                 com_port,
                 baud_rate,
                 flow_control,
                 timeout,
                 ip_address,
                 tcp_port,
                 clear_errors_on_init=True,
                 **kwargs):
        # Initialize values common to all XIP instruments
        GenericInstrument.__init__(self, serial_number, com_port, baud_rate, 8, 1, serial.PARITY_NONE, flow_control,
                                   False, timeout, ip_address, tcp_port, **kwargs)
        self.status_byte_register = StatusByteRegister
        self.standard_event_register = StandardEventRegister
        self.operation_register = None
        self.questionable_register = None
        if clear_errors_on_init:
            self.command('SYSTem:ERRor:CLEar', check_errors=False)

    def command(self, *commands, check_errors=True):
        """Send a SCPI command or multiple commands to the instrument

            Args:
                commands (str):
                    Any number of SCPI commands.

            Kwargs:
                check_errors (bool):
                    Chooses whether to query the SCPI error queue and raise errors as exceptions. True by default.

        """

        # Group all commands into a single string with SCPI delimiters.
        command_string = ";:".join(commands)

        # Pass the string to the query function if it contains a question mark.
        if check_errors:
            # Do a query which will automatically check the errors.
            self.query(command_string)
        else:
            with self.dut_lock:
                # Send command to the instrument over serial. If serial is not configured, send it over TCP.
                if self.device_serial is not None:
                    self._usb_command(command_string)
                elif self.device_tcp is not None:
                    self._tcp_command(command_string)
                else:
                    raise InstrumentException("No connections configured")

                self.logger.info('Sent SCPI command to %s: %s', self.serial_number, command_string)

    def query(self, *queries, check_errors=True):
        """Send a SCPI query or multiple queries to the instrument and return the response(s)

            Args:
                queries (str):
                    Any number of SCPI queries or commands.

            Kwargs:
                check_errors (bool):
                    Chooses whether to query the SCPI error queue and raise errors as exceptions. True by default.

            Returns:
               The instrument query response as a string.

        """

        # Group all commands and queries a single string with SCPI delimiters.
        query_string = ";:".join(queries)

        # Append the query with an additional error buffer query.
        if check_errors:
            query_string += ";:SYSTem:ERRor:ALL?"

        # Query the instrument over serial. If serial is not configured, use TCP.
        with self.dut_lock:
            if self.device_serial is not None:
                response = self._usb_query(query_string)
            elif self.device_tcp is not None:
                response = self._tcp_query(query_string)
            else:
                raise InstrumentException("No connections configured")

            self.logger.info('Sent SCPI query to %s: %s', self.serial_number, query_string)
            self.logger.info('Received SCPI response from %s: %s', self.serial_number, response)

        if check_errors:
            # Split the responses to each query, remove the last response which is to the error buffer query,
            # and check whether it contains an error
            response_list = re.split(''';(?=(?:[^'"]|'[^']*'|"[^"]*")*$)''', response)
            error_response = response_list.pop()
            self._error_check(error_response)
            response = ';'.join(response_list)

        return response

    @staticmethod
    def _error_check(error_response):
        """Evaluates the instrument response"""

        # If the error buffer returns an error, raise an exception with that includes the error.
        if "No error" not in error_response:
            raise XIPInstrumentException("SCPI command error(s): " + error_response)

    def get_status_byte(self):
        """Returns named bits of the status byte register and their values"""

        response = self.query("*STB?", check_errors=False)
        status_register = self.status_byte_register.from_integer(response)

        return status_register

    def get_service_request_enable_mask(self):
        """Returns the named bits of the status byte service request enable register.
        This register determines which bits propagate to the master summary status bit"""

        response = self.query("*SRE?", check_errors=False)
        status_register = self.status_byte_register.from_integer(response)

        return status_register

    def set_service_request_enable_mask(self, register_mask):
        """Configures values of the service request enable register bits.
        This register determines which bits propagate to the master summary bit

            Args:
                register_mask (StatusByteRegister):
                    A StatusByteRegister class object with all bits configured true or false.
        """

        integer_representation = register_mask.to_integer()
        self.command(f"*SRE {str(integer_representation)}", check_errors=False)

    def get_standard_events(self):
        """Returns the names of the standard event register bits and their values"""

        response = self.query("*ESR?", check_errors=False)
        status_register = self.standard_event_register.from_integer(response)

        return status_register

    def get_standard_event_enable_mask(self):
        """Returns the names of the standard event enable register bits and their values.
        These values determine which bits propagate to the standard event register"""

        response = self.query("*ESE?", check_errors=False)
        status_register = self.standard_event_register.from_integer(response)

        return status_register

    def set_standard_event_enable_mask(self, register_mask):
        """Configures values of the standard event enable register bits.
        These values determine which bits propagate to the standard event register

            Args:
                register_mask (StandardEventRegister):
                    A StandardEventRegister class object with all bits configured true or false.
        """

        integer_representation = register_mask.to_integer()
        self.command(f"*ESE {str(integer_representation)}", check_errors=False)

    def get_present_operation_status(self):
        """Returns the names of the operation status register bits and their values"""

        response = self.query("STATus:OPERation:CONDition?", check_errors=False)
        status_register = self.operation_register.from_integer(response)

        return status_register

    def get_operation_events(self):
        """Returns the names of operation event status register bits that are currently high.
        The event register is latching and values are reset when queried."""

        response = self.query("STATus:OPERation:EVENt?", check_errors=False)
        status_register = self.operation_register.from_integer(response)

        return status_register

    def get_operation_event_enable_mask(self):
        """Returns the names of the operation event enable register bits and their values.
        These values determine which operation bits propagate to the operation event register."""

        response = self.query("STATus:OPERation:ENABle?", check_errors=False)
        status_register = self.operation_register.from_integer(response)

        return status_register

    def set_operation_event_enable_mask(self, register_mask):
        """Configures the values of the operation event enable register bits.
        These values determine which operation bits propagate to the operation event register.

            Args:
                register_mask ([Instrument]OperationRegister):
                    An instrument specific OperationRegister class object with all bits configured true or false.
        """

        integer_representation = register_mask.to_integer()
        self.command(f"STATus:OPERation:ENABle {str(integer_representation)}", check_errors=False)

    def get_present_questionable_status(self):
        """Returns the names of the questionable status register bits and their values"""

        response = self.query("STATus:QUEStionable:CONDition?", check_errors=False)
        status_register = self.questionable_register.from_integer(response)

        return status_register

    def get_questionable_events(self):
        """Returns the names of questionable event status register bits that are currently high.
        The event register is latching and values are reset when queried."""

        response = self.query("STATus:QUEStionable:EVENt?", check_errors=False)
        status_register = self.questionable_register.from_integer(response)

        return status_register

    def get_questionable_event_enable_mask(self):
        """Returns the names of the questionable event enable register bits and their values.
        These values determine which questionable bits propagate to the questionable event register."""

        response = self.query("STATus:QUEStionable:ENABle?", check_errors=False)
        status_register = self.questionable_register.from_integer(response)

        return status_register

    def set_questionable_event_enable_mask(self, register_mask):
        """Configures the values of the questionable event enable register bits.
        These values determine which questionable bits propagate to the questionable event register.

            Args:
                register_mask ([Instrument]QuestionableRegister):
                    An instrument specific QuestionableRegister class object with all bits configured true or false.
        """

        integer_representation = self.questionable_register.to_integer(register_mask)
        self.command(f"STATus:QUEStionable:ENABle {str(integer_representation)}", check_errors=False)

    def reset_status_register_masks(self):
        """Resets status register masks to preset values"""
        self.command("STATus:PRESet", check_errors=False)

    def modify_service_request_mask(self, bit_name, value):
        """Gets the service request enable mask, changes a bit, and sets the register.

            Args:
                bit_name (str):
                    The name of the bit to modify.

                value (bool):
                    Determines whether the bit masks (false) or passes (true) the corresponding state.

        """

        mask_register = self.get_service_request_enable_mask()

        setattr(mask_register, bit_name, value)

        self.set_service_request_enable_mask(mask_register)

    def modify_standard_event_register_mask(self, bit_name, value):
        """Gets the standard event register mask, changes a bit, and sets the register

            Args:
                bit_name (str):
                    The name of the bit to modify.

                value (bool):
                    Determines whether the bit masks (false) or passes (true) the corresponding state.

        """

        mask_register = self.get_standard_event_enable_mask()

        setattr(mask_register, bit_name, value)

        self.set_standard_event_enable_mask(mask_register)

    def modify_operation_register_mask(self, bit_name, value):
        """Gets the operation condition register mask, changes a bit, and sets the register

            Args:
                bit_name (str):
                    The name of the bit to modify.

                value (bool):
                    Determines whether the bit masks (false) or passes (true) the corresponding state.

        """

        mask_register = self.get_operation_event_enable_mask()

        setattr(mask_register, bit_name, value)

        self.set_operation_event_enable_mask(mask_register)

    def modify_questionable_register_mask(self, bit_name, value):
        """Gets the questionable condition register mask, changes a bit, and sets the register

            Args:
                bit_name (str):
                    The name of the bit to modify.

                value (bool):
                    Determines whether the bit masks (false) or passes (true) the corresponding state.

        """

        mask_register = self.get_questionable_event_enable_mask()

        setattr(mask_register, bit_name, value)

        self.set_questionable_event_enable_mask(mask_register)

    def reset_measurement_settings(self):
        """Resets measurement settings to their default values."""
        self.command("SYSTEM:PRESET")

    def factory_reset(self):
        """Resets all system information such as settings, wi-fi connections, date and time, etc."""
        self.command("SYSTEM:FACTORYRESET")

    def _get_identity(self):
        return self.query('*IDN?', check_errors=False).split(',')



"""Implements functionality unique to the Lake Shore Model 643 electromagnet magnet power supply"""
import serial

from .generic_instrument import GenericInstrument


class Model643(GenericInstrument):
    """A class object representing the Lake Shore Model 643 electromagnet magnet power supply"""

    vid_pid = [(0x1FB9, 0x0601)]

    def __init__(self,
                 serial_number=None,
                 com_port=None,
                 baud_rate=57600,
                 data_bits=7,
                 stop_bits=1,
                 parity=serial.PARITY_ODD,
                 flow_control=False,
                 handshaking=False,
                 timeout=2.0,
                 ip_address=None,
                 tcp_port=7777,
                 **kwargs):

        # Call the parent init, then fill in values specific to the 121
        GenericInstrument.__init__(self, serial_number, com_port, baud_rate, data_bits, stop_bits, parity, flow_control,
                                   handshaking, timeout, ip_address, tcp_port, **kwargs)



"""Implements functionality unique to the Lake Shore Model 336 cryogenic temperature controller"""
from enum import IntEnum

from .generic_instrument import RegisterBase
from .temperature_controllers import TemperatureController, InstrumentException
from .temperature_controllers import RelayControlMode, RelayControlAlarm, InterfaceMode, HeaterError, \
    CurveFormat, CurveTemperatureCoefficient, AutotuneMode, HeaterResistance, Polarity, DiodeCurrent, \
    HeaterOutputUnits, InputSensorUnits, ControlTypes, StandardEventRegister, OperationEvent, \
    LanStatus, Interface, CurveHeader, AlarmSettings, DisplayFields, DisplayFieldsSize

Model336RelayControlMode = RelayControlMode
Model336RelayControlAlarm = RelayControlAlarm
Model336InterfaceMode = InterfaceMode
Model336HeaterError = HeaterError
Model336CurveFormat = CurveFormat
Model336CurveTemperatureCoefficients = CurveTemperatureCoefficient
Model336AutoTuneMode = AutotuneMode
Model336HeaterResistance = HeaterResistance
Model336Polarity = Polarity
Model336DiodeCurrent = DiodeCurrent
Model336HeaterOutputUnits = HeaterOutputUnits
Model336InputSensorUnits = InputSensorUnits
Model336ControlTypes = ControlTypes
Model336LanStatus = LanStatus
Model336Interface = Interface
Model336CurveHeader = CurveHeader
Model336AlarmSettings = AlarmSettings
Model336DisplayFields = DisplayFields
Model336DisplayFieldsSize = DisplayFieldsSize
Model336StandardEventRegister = StandardEventRegister
Model336OperationEvent = OperationEvent


class Model336InputChannel(IntEnum):
    """Enumeration where "NONE" is an option for sensor input"""
    NONE = 0
    CHANNEL_A = 1
    CHANNEL_B = 2
    CHANNEL_C = 3
    CHANNEL_D = 4
    CHANNEL_D2 = 5
    CHANNEL_D3 = 6
    CHANNEL_D4 = 7
    CHANNEL_D5 = 8


class Model336DisplaySetupMode(IntEnum):
    """Front panel display setup enum"""
    INPUT_A = 0
    INPUT_B = 1
    INPUT_C = 2
    INPUT_D = 3
    CUSTOM = 4
    FOUR_LOOP = 5
    ALL_INPUTS = 6
    INPUT_D2 = 7
    INPUT_D3 = 8
    INPUT_D4 = 9
    INPUT_D5 = 10


class Model336InputSensorType(IntEnum):
    """Sensor type enumeration. THERMOCOUPLE is only valid with the 3060 option,
        CAPACITANCE is only valid with the 3061 option
    """
    DISABLED = 0
    DIODE = 1
    PLATINUM_RTD = 2
    NTC_RTD = 3
    THERMOCOUPLE = 4
    CAPACITANCE = 5


class Model336DiodeRange(IntEnum):
    """Diode voltage range enumeration"""
    TWO_POINT_FIVE_VOLTS = 0
    TEN_VOLTS = 1


class Model336RTDRange(IntEnum):
    """RTD resistance range enumeration. THIRTY_THOUSAND_OHM and
        ONE_HUNDRED_THOUSAND_OHM are only valid for NTC RTDs
    """
    TEN_OHM = 0
    THIRTY_OHM = 1
    HUNDRED_OHM = 2
    THREE_HUNDRED_OHM = 3
    ONE_THOUSAND_OHM = 4
    THREE_THOUSAND_OHM = 5
    TEN_THOUSAND_OHM = 6
    THIRTY_THOUSAND_OHM = 7
    ONE_HUNDRED_THOUSAND_OHM = 8


class Model336ThermocoupleRange(IntEnum):
    """Thermocouple range enumeration"""
    FIFTY_MILLIVOLT = 0


class Model336InputSensorSettings:
    """Class object used in the get/set_input_sensor methods"""

    def __init__(self, sensor_type, autorange_enable, compensation, units, input_range=None):
        """Constructor for the InputSensorSettings class

            Args:
                sensor_type (Model336InputSensorType):
                    * Specifies input sensor type

                autorange_enable (bool):
                    * Specifies autoranging
                    * False = off and True = on

                compensation (bool):
                    * Specifies input compensation
                    * False = off and True = on

                units (Model336InputSensorUnits):
                    * Specifies the preferred units parameter for sensor readings and for the control setpoint

                input_range (IntEnum)
                    * Specifies input range if autorange_enable is false
                    * See IntEnum classes:
                        * Model336DiodeRange
                        * Model336RTDRange
                        * Model336ThermocoupleRange

        """
        self.sensor_type = sensor_type
        self.autorange_enable = autorange_enable
        self.compensation = compensation
        self.units = units
        self.input_range = input_range


class Model336HeaterOutputMode(IntEnum):
    """Control loop enumeration"""
    OFF = 0
    CLOSED_LOOP = 1
    ZONE = 2
    OPEN_LOOP = 3
    MONITOR_OUT = 4
    WARMUP_SUPPLY = 5


class Model336HeaterRange(IntEnum):
    """Current mode heater enumerations"""
    OFF = 0
    LOW = 1
    MEDIUM = 2
    HIGH = 3


class Model336HeaterVoltageRange(IntEnum):
    """Voltage mode heater enumerations"""
    VOLTAGE_OFF = 0
    VOLTAGE_ON = 1


class Model336DisplayUnits(IntEnum):
    """Panel display units enumeration"""
    KELVIN = 1
    CELSIUS = 2
    SENSOR_UNITS = 3
    MINIMUM_DATA = 4
    MAXIMUM_DATA = 5
    SENSOR_NAME = 6


class Model336ControlLoopZoneSettings:
    """Control loop configuarion for a particular heater output and zone"""

    def __init__(self, upper_bound, proportional, integral, derivative, manual_out_value, heater_range, channel, rate):
        """Constructor

            Args:
                upper_bound (float):
                    * Specifies the upper Setpoint boundary of this zone in kelvin

                proportional (float):
                    * Specifies the proportional gain for this zone
                    * 0.1 to 1000

                integral (float):
                    * Specifies the integral gain for this zone
                    * 0.1 to 1000

                derivative (float):
                    * Specifies the derivative gain for this zone
                    * 0 to 200 %

                manual_out_value (float):
                    * Specifies the manual output for this zone
                    * 0 to 100 %

                heater_range (Model336HeaterRange):
                    * Specifies the heater range for this zone
                    * See Model336HeaterRange IntEnum class

                channel (Model336InputChannel):
                    * See Model336InputChannel IntEnum class
                    * Passing the NONE member will use the previously assigned sensor

                rate (float):
                    * Specifies the ramp rate for this zone
                    * 0 - 100 K/min

        """

        self.upper_bound = upper_bound
        self.proportional = proportional
        self.integral = integral
        self.derivative = derivative
        self.manual_out_value = manual_out_value
        self.heater_range = heater_range
        self.channel = channel
        self.rate = rate


class Model336StatusByteRegister(RegisterBase):
    """Class object representing the status byte register LSB to MSB"""

    bit_names = [
        "",
        "",
        "",
        "",
        "message_available_summary_bit",
        "event_status_summary_bit",
        "service_request",
        "operation_summary_bit"
    ]

    def __init__(self,
                 message_available_summary_bit,
                 event_status_summary_bit,
                 service_request,
                 operation_summary_bit):
        self.message_available_summary_bit = message_available_summary_bit
        self.event_status_summary_bit = event_status_summary_bit
        self.service_request = service_request
        self.operation_summary_bit = operation_summary_bit


class Model336ServiceRequestEnable(RegisterBase):
    """Class object representing the service request enable register LSB to MSB"""

    bit_names = [
        "",
        "",
        "",
        "",
        "message_available_summary_bit",
        "event_status_summary_bit",
        "",
        "operation_summary_bit"
    ]

    def __init__(self,
                 message_available_summary_bit,
                 event_status_summary_bit,
                 operation_summary_bit):
        self.message_available_summary_bit = message_available_summary_bit
        self.event_status_summary_bit = event_status_summary_bit
        self.operation_summary_bit = operation_summary_bit


class Model336InputReadingStatus(RegisterBase):
    """Class object representing the input staus flag bits"""

    bit_names = [
        "invalid_reading",
        "",
        "",
        "",
        "temp_underrange",
        "temp_overrange",
        "sensor_units_zero",
        "sensor_units_overrange"
    ]

    def __init__(self, invalid_reading, temp_underrange, temp_overrange, sensor_units_zero, sensor_units_overrange):
        self.invalid_reading = invalid_reading
        self.temp_underrange = temp_underrange
        self.temp_overrange = temp_overrange
        self.sensor_units_zero = sensor_units_zero
        self.sensor_units_overrange = sensor_units_overrange


class Model336(TemperatureController):
    """A class object representing the Lake Shore Model 336 cryogenic temperature controller"""

    vid_pid = [(0x1FB9, 0x0301)]

    # Initiate enum types for temperature controllers class
    _input_channel_enum = Model336InputChannel
    _display_units_enum = Model336DisplayUnits

    # Initiate instrument specific registers
    status_byte_register = Model336StatusByteRegister
    service_request_enable = Model336ServiceRequestEnable

    def __init__(self,
                 serial_number=None,
                 com_port=None,
                 timeout=2.0,
                 ip_address=None,
                 tcp_port=7777,
                 **kwargs):

        # Call the parent init, then fill in values specific to the 336
        TemperatureController.__init__(self, serial_number, com_port, 57600, timeout, ip_address,
                                       tcp_port, **kwargs)

    # Alias specific temperature controller methods
    get_analog_output_percentage = TemperatureController._get_analog_output_percentage
    set_autotune = TemperatureController._set_autotune
    set_contrast_level = TemperatureController._set_contrast_level
    get_contrast_level = TemperatureController._get_contrast_level
    get_operation_condition = TemperatureController._get_operation_condition
    get_operation_event_enable = TemperatureController._get_operation_event_enable
    set_operation_event_enable = TemperatureController._set_operation_event_enable
    get_operation_event = TemperatureController._get_operation_event
    get_thermocouple_junction_temp = TemperatureController._get_thermocouple_junction_temp
    set_soft_cal_curve_dt_470 = TemperatureController._set_soft_cal_curve_dt_470
    set_soft_cal_curve_pt_100 = TemperatureController._set_soft_cal_curve_pt_100
    set_soft_cal_curve_pt_1000 = TemperatureController._set_soft_cal_curve_pt_1000
    set_filter = TemperatureController._set_filter
    get_filter = TemperatureController._get_filter
    set_network_settings = TemperatureController._set_network_settings
    get_network_settings = TemperatureController._get_network_settings
    get_network_configuration = TemperatureController._get_network_configuration
    set_website_login = TemperatureController._set_website_login
    get_website_login = TemperatureController._get_website_login
    get_celsius_reading = TemperatureController._get_celsius_reading
    set_interface = TemperatureController._set_interface
    get_interface = TemperatureController._get_interface
    get_tuning_control_status = TemperatureController._get_tuning_control_status
    set_diode_excitation_current = TemperatureController._set_diode_excitation_current
    get_diode_excitation_current = TemperatureController._get_diode_excitation_current

    def set_monitor_output_heater(self, output, channel, units, high_value, low_value, polarity):
        """Configures a voltage-controlled output. Use the set_heater_output_mode
        command to set the output mode to Monitor Out.

            Args:
                output (int):
                    * Voltage-controlled output to configure
                    * 3 or 4

                channel (Model336InputChannel):
                    * Specifies which sensor input to monitor
                    * A member of the Model336InputChannel IntEnum class

                units (Model336InputSensorUnits):
                    * Specifies the units on which to base the output voltage
                    * A member of the Model336InputSensorUnits IntEnum class

                high_value (float):
                    * Represents the data at which the Monitor Out reaches +100% output
                    * Entered in the units designated by the <units> argument

                low_value (float):
                    * Represents the data at which the analog output reaches -100% output if bipolar,
                    * or 0% outputif unipolar. Entered in the units designated by the <units> argument

                polarity (Model336Polarity):
                    * Specifies whether the output voltage is unipolar or bipolar
                    * Member of the Model336Polarity IntEnum class

        """
        command_string = f"ANALOG {output},{channel},{units},{high_value},{low_value},{polarity}"
        self.command(command_string)

    def get_monitor_output_heater(self, output):
        """Used to obtain all monitor out parameters for a specific output

            Args:
                output (int):
                    * Voltage-controlled output to configure
                    * 3 or 4

            Return:
                (dict):
                    * See set_monitor_output_heater arguments

        """
        response = self.query(f"ANALOG? {output}").split(",")
        return {"channel": Model336InputChannel(int(response[0])),
                "units": Model336InputSensorUnits(int(response[1])),
                "high_value": float(response[2]),
                "low_value": float(response[3]),
                "polarity": Model336Polarity(int(response[4]))}

    def set_display_setup(self, mode, num_fields="", displayed_output=""):
        """Sets the display mode

            Args:
                mode (Model336DisplaySetupMode):
                    * Member of Model336DisplaySetupMode IntEnum class
                    * Specifies display mode for default and 3062 options

                num_fields (IntEnum)
                    * When mode is set to custom, specifies the number of fields
                        * Member of Model336DisplayFields
                    * When mode is set to all inputs, specifies size of readings
                        * Member of Model336DisplayFieldsSize

                displayed_output (int):
                    * Configures the bottom half of the custom display screen
                    * Only required if mode is set to CUSTOM
                        * Output: 1 - 4

        """
        if mode == Model336DisplaySetupMode.CUSTOM:
            if not isinstance(num_fields, Model336DisplayFields):
                raise InstrumentException("num_fields argument must be of type \"Model336DisplaySetupCustom\"")
        elif mode == Model336DisplaySetupMode.ALL_INPUTS:
            if not isinstance(num_fields, Model336DisplayFieldsSize):
                raise InstrumentException("num_fields argument must be of type \"Model336DisplaySetupAllInputs\"")

        command_string = f"DISPLAY {mode},{num_fields},{displayed_output}"
        self.command(command_string)

    def get_display_setup(self):
        """Returns the display mode

            Return:
                (dict)
                    * See set_display_setup method arguments
                    * Keys: "mode", "num_fields", "displayed_output"

        """
        display_setup_response = self.query("DISPLAY?").split(",")
        mode = Model336DisplaySetupMode(int(display_setup_response[0]))
        if mode == Model336DisplaySetupMode.CUSTOM:
            num_fields = Model336DisplayFields(int(display_setup_response[1]))
            displayed_output = int(display_setup_response[2])
        elif mode == Model336DisplaySetupMode.ALL_INPUTS:
            num_fields = Model336DisplayFieldsSize(int(display_setup_response[1]))
            displayed_output = None
        else:
            num_fields = None
            displayed_output = None

        return {"mode": mode,
                "num_fields": num_fields,
                "displayed_output": displayed_output}

    def set_heater_setup(self, output, heater_resistance, max_current, heater_output):
        """Method to configure the heaters

            Args:
                output (int):
                    * Specifies which heater output to configure
                    * 1 or 2

                heater_resistance (Model336HeaterResistance):
                    * Member of Model336HeaterResistance IntEnum class

                max_current (float):
                    * User defined maximum output current (see table 4-11 for max current and resistance relationships)

                heater_output (Model336HeaterOutputUnits):
                    * Specifies whether the heater output displays in current or power
                    * Member of Model336HeaterOutputUnits IntEnum class

        """
        self.command(f"HTRSET {output},{heater_resistance},0,{max_current},{heater_output}")

    def get_heater_setup(self, heater_output):
        """Returns the heater configuration status.

            Args:
                heater_output (int):
                    * Specifies which heater output to configure
                    * 1 or 2

            Return:
                (dict):
                    * See set_heater_setup arguments
                    * Keys:
                        * heater_resistance
                        * max_current
                        * output_display_mode

        """
        heater_setup = self.query(f"HTRSET? {heater_output}").split(",")
        if int(heater_setup[1]) == 0:
            max_current = float(heater_setup[2])
        else:
            preset_currents = ["USER", 0.707, 1.0, 1.141, 2.0]
            current_index = int(heater_setup[1])
            max_current = preset_currents[current_index]

        return {"heater_resistance": Model336HeaterResistance(int(heater_setup[0])),
                "max_current": max_current,
                "output_display_mode": Model336HeaterOutputUnits(int(heater_setup[3]))}

    def set_input_sensor(self, channel, sensor_parameters):
        """Sets the sensor type and associated parameters.

            Args:
                channel (str):
                    * Specifies input to configure
                        * "A" - "D"
                    * 3062 option:
                        * "D1" - "D5"

                sensor_parameters (Model336InputSensorSettings):
                    * See Model336InputSensorSettings class

        """
        autorange_enable = sensor_parameters.autorange_enable
        if autorange_enable:
            input_range = 0
        else:
            input_range = sensor_parameters.input_range

        command_string = (f"INTYPE {channel},{sensor_parameters.sensor_type},{int(sensor_parameters.autorange_enable)}," +
                            f"{input_range},{int(sensor_parameters.compensation)},{sensor_parameters.units}")
        self.command(command_string)

    def get_input_sensor(self, channel):
        """Returns the sensor type and associated parameters.

            Args:
                channel (str):
                    * Specifies sensor input to configure
                    * "A" or "B"

            Return:
                (Model336InputSensorSettings):
                    * See Model336InputSensorSettings class

        """
        sensor_config = self.query(f"INTYPE? {channel}").split(",")
        sensor_type = Model336InputSensorType(int(sensor_config[0]))
        autorange_enable = bool(int(sensor_config[1]))
        if autorange_enable:
            input_range = 0
        else:
            if sensor_type == Model336InputSensorType.DIODE:
                input_range = Model336DiodeRange(int(sensor_config[2]))
            elif sensor_type in (Model336InputSensorType.PLATINUM_RTD, Model336InputSensorType.NTC_RTD):
                input_range = Model336RTDRange(int(sensor_config[2]))
            elif sensor_type == Model336InputSensorType.THERMOCOUPLE:
                input_range = Model336ThermocoupleRange(int(sensor_config[2]))
            else:
                input_range = int(sensor_config[2])

        return Model336InputSensorSettings(sensor_type,
                                           bool(int(sensor_config[1])),
                                           bool(int(sensor_config[3])),
                                           Model336InputSensorUnits(int(sensor_config[4])),
                                           input_range)

    def get_all_kelvin_reading(self):
        """Returns the temperature value in kelvin of all channels.

            Return:
                (list: float)
                    * [channel_A, channel_B, channel_C, channel_D]

        """
        kelvin_reading = self.query("KRDG? 0").split(",")
        return [float(channel) for channel in kelvin_reading]

    def set_heater_output_mode(self, output, mode, channel, powerup_enable=False):
        """Configures the heater output mode.

            Args:
                output (int):
                    * Specifies which output to configure
                    * 1 - 4

                mode (Model336HeaterOutputMode):
                    * Member of Model336HeaterOutputMode IntEnum class
                    * Specifies the control mode

                channel (Model336InputChannel):
                    * Model336InputChannel IntEnum class
                    * Specifies which input to use for control

                powerup_enable (bool)
                    * Specifies whether the output remains on (True)
                    * or shuts off after power cycle (False)

        """
        command_string = f"OUTMODE {output},{mode},{channel},{int(powerup_enable)}"
        self.command(command_string)

    def get_heater_output_mode(self, output):
        """Returns the heater output mode for a given output and whether powerup is enabled.

            Args:
                output (int):
                    * Specifies which output to retrieve
                    * 1 - 4

            Return:
                (dict):
                    * See set_heater_output_mode method arguments
                    * Keys:
                        * mode
                        * channel
                        * powerup_enable

        """
        outmode = self.query(f"OUTMODE? {output}").split(",")
        return {"mode": Model336HeaterOutputMode(int(outmode[0])),
                "channel": Model336InputChannel(int(outmode[1])),
                "powerup_enable": bool(int(outmode[2]))}

    def set_heater_range(self, output, heater_range):
        """Sets the heater range for a particular output.
        The range setting has no effect if an output is in
        the Off mode, and does not apply to an output in Monitor
        Out mode. An output in Monitor Out mode is always on.

            Args:
                output (int):
                    * Specifies which output to configure
                    * 1 - 4

                heater_range (IntEnum):
                    * For Outputs 1 and 2:
                        * Member of Model336HeaterRange IntEnum class
                    * For Outputs 3 and 4:
                        * Model336HeaterVoltageRange IntEnum class

        """
        self.command(f"RANGE {output},{heater_range}")

    def get_heater_range(self, output):
        """Returns the heater range for a particular output.

            Args:
                output (int):
                    * Specifies which output to query (1 or 2)

            Return:
                (IntEnum):
                    * For Outputs 1 and 2:
                        * Member of Model336HeaterRange IntEnum class
                    * For Outputs 3 and 4:
                        * Member of Model336HeaterVoltageRange IntEnum class

        """
        heater_range = int(self.query(f"RANGE? {output}"))

        if output in (3, 4):
            heater_range = Model336HeaterVoltageRange(heater_range)
        else:
            heater_range = Model336HeaterRange(heater_range)

        return heater_range

    def all_heaters_off(self):
        """Recreates the front panel safety feature of shutting off all heaters"""
        self.command("RANGE 1,0;RANGE 2,0;RANGE 3,0;RANGE 4,0")

    def get_input_reading_status(self, channel):
        """Retruns the state of the input status flag bits.

            Args:
                channel (str):
                    * Specifies which channel to query
                    * "A" - "D"
                    * "D1" - "D5" for 3062 option

            Return:
                (Model336InputReadingStatus):
                    * Boolean representation of each bit in the input status flag register

        """
        response = int(self.query(f"RDGST? {channel}"))
        return Model336InputReadingStatus.from_integer(response)

    def get_all_sensor_reading(self):
        """Returns the sensor unit reading of all channels

            Return:
                (list: float):
                    * [channel_A, channel_B, channel_C, channel_D]

        """
        sensor_reading = self.query("SRDG? 0").split(",")
        return [float(channel) for channel in sensor_reading]

    def set_warmup_supply_parameter(self, output, control, percentage):
        """Warmup mode applies only to voltage heater outputs 3 and 4. The Output mode and Control
            Input parameters must be configured using the set_monitor_out_parameters() method.

            Args:
                output (int):
                    * Specifies which output to configure
                    * 3 or 4

                control (Model336ControlTypes):
                    * Member of the Model336ControlTypes IntEnum class

                percentage (float):
                    * Specifies the percentage of full scale (10 V) Monitor Out
                    * voltage to apply to turn on the external power supply
                        * A value of 50.5 translates to a 50.5 percent output voltage

        """
        command_string = f"WARMUP {output},{control},{percentage}"
        self.command(command_string)

    def get_warmup_supply_parameter(self, output):
        """Returns the warmup supply configuration for a particular output.

            Args:
                output (int):
                    * Specifies which analog voltage heater output to retrieve
                    * 3 or 4

            Return:
                (dict):
                    * See set_warmup_supply_parameter method arguments

        """
        warmup_supply = self.query(f"WARMUP? {output}").split(",")
        return {"control": Model336ControlTypes(int(warmup_supply[0])),
                "percentage": float(warmup_supply[1])}

    def set_control_loop_zone_table(self, output, zone, control_loop_zone):
        """Configures the output zone parameters.

            Args:
                output (int):
                    * Specifies which analog voltage heater output to configure
                    * 1 or 2

                zone (int):
                    * Specifies which zone in the table to configure
                    * 1 to 10

                control_loop_zone (Model336ControlLoopZoneSettings):
                    * See Model336ControlLoopZoneSettings class

        """
        command_string = (f"ZONE {output},{zone},{control_loop_zone.upper_bound},{control_loop_zone.proportional}," +
                            f"{control_loop_zone.integral},{control_loop_zone.derivative}," +
                            f"{control_loop_zone.manual_out_value},{control_loop_zone.heater_range}," +
                            f"{control_loop_zone.channel},{control_loop_zone.rate}")
        self.command(command_string)

    def get_control_loop_zone_table(self, output, zone):
        """Returns a list of zone control parameters for a selected output and zone.

            Args:
                output (int):
                    * Specifies which heater output to query
                    * 1 or 2

                zone (int):
                    * Specifies which zone in the table to query
                    * 1 to 10

            Return:
                (Model336ControlLoopZoneSettings):
                    * See Model336ControlLoopZoneSettings class

        """
        zone_parameters = self.query(f"ZONE? {output},{zone}").split(",")
        return Model336ControlLoopZoneSettings(float(zone_parameters[0]),
                                               float(zone_parameters[1]),
                                               float(zone_parameters[2]),
                                               float(zone_parameters[3]),
                                               float(zone_parameters[4]),
                                               Model336HeaterRange(int(zone_parameters[5])),
                                               Model336InputChannel(int(zone_parameters[6])),
                                               float(zone_parameters[7]))

    def _autotune_error(self):
        """Method to raise an excecption if autotune error has occured"""

        tuning_status = self.query("TUNEST?").split(",")

        if bool(int(tuning_status[2])):
            raise InstrumentException("An autotune error is present")


__all__ = ['Model336RelayControlMode', 'Model336RelayControlAlarm', 'Model336InterfaceMode', 'Model336HeaterError',
           'Model336CurveFormat', 'Model336CurveTemperatureCoefficients', 'Model336AutoTuneMode',
           'Model336HeaterResistance', 'Model336Polarity', 'Model336DiodeCurrent', 'Model336HeaterOutputUnits',
           'Model336InputSensorUnits', 'Model336ControlTypes', 'Model336LanStatus', 'Model336Interface',
           'Model336CurveHeader', 'Model336AlarmSettings', 'Model336StandardEventRegister', 'Model336OperationEvent',
           'Model336InputChannel', 'Model336DisplaySetupMode', 'Model336DisplayFields',
           'Model336DisplayFieldsSize', 'Model336InputSensorType', 'Model336DiodeRange', 'Model336RTDRange',
           'Model336ThermocoupleRange', 'Model336InputSensorSettings', 'Model336HeaterOutputMode',
           'Model336HeaterRange', 'Model336HeaterVoltageRange', 'Model336DisplayUnits',
           'Model336ControlLoopZoneSettings', 'Model336StatusByteRegister', 'Model336ServiceRequestEnable',
           'Model336InputReadingStatus', 'AlarmSettings', 'Model336']
